"""Code for loading up an override file.

The override file provides implementations of functions where the code
generator could not do its job correctly.

This is a simple rip-off of the override script used in PyGTK.
"""

import re
from typing import List

OVERRIDE_RE = re.compile(
    r"^override\s+(?P<name>[\w.]+)(?:\((?P<derived>[^)]+)\))?\s*(?::\s*(?P<type_hint>[\w\s\[\],\"| ]+))?$"
)


class Overrides:
    def __init__(self, filename=None):
        self.overrides = {}
        self.header = ""
        if filename:
            with open(filename, encoding="utf-8") as fp:
                self.read_overrides(fp)

    def read_overrides(self, fp):
        """Read a file and return a dictionary of overridden properties and
        their implementation.

        An override file ahs the form: override <property>
        <implementation> %%
        """
        # read all the components of the file ...
        # bufs contains a list of (lines, line_number) pairs.
        bufs = []
        line_number = 1
        lines: List[str] = []
        line = fp.readline()
        linenum = 1
        while line:
            if line in ("%%\n", "%%"):
                if lines:
                    bufs.append((lines.copy(), line_number))
                line_number = linenum + 1
                lines = []
            else:
                lines.append(line)
            line = fp.readline()
            linenum = linenum + 1
        if lines:
            bufs.append((lines.copy(), line_number))

        if not bufs:
            return

        # Parse the parts of the file
        for lines, line_number in bufs:
            line = lines[0]
            rest = lines[1:]
            words = line.split()

            if words[0] == "override":
                m = OVERRIDE_RE.match(line.strip())
                if not m:
                    raise Exception(f"Could not parse override line '{line.strip()}'")
                func = m.group("name")
                derived = m.group("derived")
                deps = tuple(map(str.strip, derived.split(","))) if derived else ()
                type_hint = m.group("type_hint") or "Any"
                self.overrides[func] = (
                    deps,
                    type_hint,
                    "".join(rest),
                    f"{line_number:d}: {line}",
                )
            elif words[0] == "header":
                assert not self.header
                self.header = "".join(rest)
            elif words[0] != "comment":
                print(f"Unknown word: '{words[0]}', line {line_number:d}")
                raise SystemExit

    def has_override(self, key):
        return key in self.overrides

    def get_override(self, key):
        """Write override data for 'key' to a file referred to by 'fp'."""
        _deps, _type_hint, data, line = self.overrides.get(key, ((), None, None, None))
        return f"# {line}{data}" if data else None

    def get_type(self, key):
        return self.overrides.get(key, (None, "Any"))[1]

    def derives(self, key):
        return self.overrides.get(key, ((),))[0]



from gi.repository import Gdk, Gio, GLib, Gtk

from gaphor import UML
from gaphor.core import transactional
from gaphor.diagram.propertypages import (
    LabelValue,
    PropertyPageBase,
    new_resource_builder,
)
from gaphor.i18n import translated_ui_string

new_builder = new_resource_builder("gaphor.UML")


class TypedElementPropertyPage(PropertyPageBase):
    order = 31

    def __init__(self, item):
        super().__init__()
        self.item = item

        assert (not item.subject) or isinstance(
            self.typed_element, UML.TypedElement
        ), item.subject

    @property
    def typed_element(self):
        return self.item.subject

    def construct(self):
        if not self.typed_element:
            return

        typed_element = self.typed_element

        builder = new_builder(
            "typed-element-editor",
            signals={
                "show-type-changed": (self._on_show_type_change,),
            },
        )

        dropdown = builder.get_object("element-type")
        model = list_of_classifiers(typed_element.model, UML.Classifier)
        dropdown.set_model(model)

        if typed_element.type:
            dropdown.set_selected(
                next(
                    n for n, lv in enumerate(model) if lv.value == typed_element.type.id
                )
            )

        dropdown.connect("notify::selected", self._on_property_type_changed)

        if hasattr(self.item, "show_type"):
            show_type = builder.get_object("show-type")
            show_type.set_active(self.item.show_type)
        else:
            builder.get_object("type-toggle-box").unparent()

        return builder.get_object("typed-element-editor")

    @transactional
    def _on_property_type_changed(self, dropdown, _pspec):
        typed_element = self.typed_element
        if id := dropdown.get_selected_item().value:
            element = typed_element.model.lookup(id)
            assert isinstance(element, UML.Type)
            typed_element.type = element
        else:
            del typed_element.type

    @transactional
    def _on_show_type_change(self, button, _gparam):
        self.item.show_type = button.get_active()


def list_of_classifiers(element_factory, required_type):
    model = Gio.ListStore.new(LabelValue)
    model.append(LabelValue("", None))
    for c in sorted(
        (c for c in element_factory.select(required_type) if c.name),
        key=lambda c: c.name or "",
    ):
        model.append(LabelValue(c.name, c.id))
    return model


def list_item_factory(
    ui_filename, klass, attribute, placeholder_text="", signal_handlers=None
):
    ui_string = translated_ui_string("gaphor.UML", ui_filename).format(
        gtype_name=klass.__gtype__.name,
        attribute=attribute.name,
        placeholder_text=placeholder_text,
    )

    return Gtk.BuilderListItemFactory.new_from_bytes(
        Gtk.Builder.BuilderScope(signal_handlers),
        GLib.Bytes.new(ui_string.encode("utf-8")),
    )


def create_list_store(type, elements, create_element):
    store = Gio.ListStore.new(type)

    for attr in elements:
        store.append(create_element(attr))

    store.append(create_element(None))

    return store


def update_list_store(
    store: Gio.ListStore, get_element, elements, create_element
) -> Gio.ListStore:
    n = 0
    for element in elements:
        if (item := store.get_item(n)) is not None and element is not get_element(item):
            store.remove(n)
            store.insert(n, create_element(element))
        n += 1

    while store.get_n_items() > n:
        store.remove(store.get_n_items() - 1)

    if (
        not store.get_n_items()
        or get_element(store.get_item(store.get_n_items() - 1)) is not None
    ):
        store.append(create_element(None))
    return store


def text_field_handlers(model_field: str):
    def on_done_editing(list_item, should_commit):
        text = list_item.get_child()
        if should_commit:
            setattr(list_item.get_item(), model_field, text.editable_text)

    return {
        "on_done_editing": on_done_editing,
    }


def check_button_handlers(model_field: str):
    def on_toggled(list_item):
        button = list_item.get_child()
        setattr(list_item.get_item(), model_field, button.get_active())

    return {
        "on_toggled": on_toggled,
    }


@transactional
def list_view_activated(list_view, _row):
    """Default action for activation: start editing."""
    selection = list_view.get_model()
    item = selection.get_selected_item()

    item.start_editing()


@transactional
def list_view_key_handler(ctrl, keyval, _keycode, state):
    """Handle keyboard shortcuts in a list view in the property editor.

    Attach to ``Gtk.EventControllerKey:key-pressed`` signals.
    """
    list_view = ctrl.get_widget()
    selection = list_view.get_model()
    item = selection.get_selected_item()

    if keyval in (Gdk.KEY_F2,):
        item.start_editing()
        return True

    if not item or item.empty():
        return False

    if keyval in (Gdk.KEY_Delete, Gdk.KEY_BackSpace) and not state & (
        Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK
    ):
        item.unlink()
        return True

    elif keyval in (Gdk.KEY_equal, Gdk.KEY_plus, Gdk.KEY_minus, Gdk.KEY_underscore):
        pos = selection.get_selected()
        swap_pos = pos + 1 if keyval in (Gdk.KEY_equal, Gdk.KEY_plus) else pos - 1
        if not 0 <= swap_pos < selection.get_n_items():
            return False

        other = selection.get_item(swap_pos)
        if not other or other.empty():
            return False

        if item.swap(item, other):
            selection.set_selected(swap_pos)
        return True

    return False



"""PictureItem diagram item."""

import base64
import io

import cairo
from PIL import Image

from gaphor.core.modeling import Picture
from gaphor.diagram.presentation import ElementPresentation
from gaphor.diagram.shapes import Box, IconBox
from gaphor.diagram.support import represents


@represents(Picture)
class PictureItem(ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.width = 200
        self.height = 200

        self.shape = IconBox(Box(draw=self.draw_image))

        self.watch("subject[Picture].content")

    def create_default_surface(self):
        width = int(self.width)
        height = int(self.height)

        surface = cairo.ImageSurface(cairo.FORMAT_RGB24, width, height)
        ctx = cairo.Context(surface)

        # Draw a white background

        ctx.set_source_rgb(1, 1, 1)
        ctx.rectangle(0, 0, width, height)
        ctx.fill()

        ctx.set_source_rgb(0, 0, 0)
        ctx.set_line_width(1)

        # Draw a border

        ctx.rectangle(0, 0, width - 1, height - 1)
        ctx.stroke()

        # Draw an X

        ctx.move_to(0, 0)
        ctx.line_to(width - 1, height - 1)
        ctx.stroke()

        ctx.move_to(width - 1, 0)
        ctx.line_to(0, height - 1)
        ctx.stroke()

        return 1.0, surface

    def create_content_surface(self):
        base64_img_bytes = self.subject.content.encode("ascii")
        image_data = base64.decodebytes(base64_img_bytes)
        image = Image.open(io.BytesIO(image_data))
        surface = self._from_pil(image)

        surface_width = surface.get_width()
        surface_height = surface.get_height()

        width_ratio = float(self.width) / float(surface_width)
        height_ratio = float(self.height) / float(surface_height)
        scale_xy = min(height_ratio, width_ratio)

        return scale_xy, surface

    def draw_image(self, box, context, bounding_box):
        scale_xy = 1.0
        surface = None

        if not self.subject or not self.subject.content:
            scale_xy, surface = self.create_default_surface()
        else:
            scale_xy, surface = self.create_content_surface()

        cr = context.cairo
        cr.save()
        cr.scale(scale_xy, scale_xy)
        cr.set_source_surface(surface, 0, 0)
        cr.paint()
        cr.restore()

        context_width = surface.get_width() * scale_xy
        context_height = surface.get_height() * scale_xy
        self.width = min(self.width, context_width)
        self.height = min(self.height, context_height)

    def _from_pil(
        self,
        im: Image,
        alpha: float = 1.0,
        format: cairo.Format = cairo.FORMAT_ARGB32,
    ) -> cairo.ImageSurface:
        """Create a new Cairo image surface from a Pillow Image

        Args:
            im (Image): Pillow Image
            alpha (float): 0..1 alpha to add to non-alpha images
            format (cairo.Format): Pixel format for output surface

        Returns:
            cairo.ImageSurface: The image surface representing the pillow image
        """
        assert format in (
            cairo.FORMAT_RGB24,
            cairo.FORMAT_ARGB32,
        ), f"Unsupported pixel format: {format}"
        if "A" not in im.getbands():
            im.putalpha(int(alpha * 256.0))
        arr = bytearray(im.tobytes("raw", "BGRa"))
        return cairo.ImageSurface.create_for_data(arr, format, im.width, im.height)



from __future__ import annotations

import logging

from gaphas.geometry import Rectangle
from gaphas.guide import GuidedItemMoveMixin
from gaphas.item import NW, SE, Item
from gaphas.move import ItemMove
from gaphas.move import Move as MoveAspect
from gaphas.tool.itemtool import item_at_point
from gaphas.view import GtkView
from gi.repository import Gtk

from gaphor.core.modeling import Element
from gaphor.diagram.connectors import can_connect
from gaphor.diagram.group import can_group, group, ungroup
from gaphor.diagram.presentation import (
    ElementPresentation,
    LinePresentation,
    Presentation,
)
from gaphor.UML.recipes import owner_package

log = logging.getLogger(__name__)


def drop_zone_tool(
    item_class: type[Presentation], subject_class: type[Element] | None
) -> Gtk.EventController:
    ctrl = Gtk.EventControllerMotion.new()
    ctrl.connect("motion", on_motion, item_class, subject_class)
    return ctrl


def on_motion(
    controller,
    x,
    y,
    item_class: type[Presentation],
    subject_class: type[Element] | None,
):
    view: GtkView = controller.get_widget()
    model = view.model

    try:
        parent = next(item_at_point(view, (x, y)), None)
    except KeyError:
        parent = None

    if parent and subject_class:
        view.selection.dropzone_item = (
            parent
            if can_group(parent.subject, subject_class)
            or can_connect(parent, item_class)  # type: ignore[arg-type]
            else None
        )
        model.request_update(parent)
    else:
        if view.selection.dropzone_item:
            model.request_update(view.selection.dropzone_item)
        view.selection.dropzone_item = None


class DropZoneMoveMixin:
    view: GtkView
    item: Item

    def start_move(self, pos):
        super().start_move(pos)  # type: ignore[misc]
        if self.item.parent:
            self.view.selection.dropzone_item = self.item.parent

    def move(self, pos):
        """Move the item.

        x and y are in view coordinates.
        """
        super().move(pos)  # type: ignore[misc]
        item = self.item
        view = self.view
        x, y = pos

        over_item = next(
            item_at_point(view, (x, y), exclude=view.selection.selected_items), None
        )

        if not over_item:
            view.selection.dropzone_item = None
            return

        if item.subject and can_group(over_item.subject, item.subject):
            view.selection.dropzone_item = over_item
            over_item.request_update()

    def stop_move(self, pos):
        """Motion stops: drop!"""
        super().stop_move(pos)  # type: ignore[misc]
        item = self.item
        view = self.view
        old_parent = item.parent
        new_parent = view.selection.dropzone_item

        if new_parent is old_parent:
            if old_parent is not None:
                old_parent.request_update()
            return

        if old_parent and ungroup(old_parent.subject, item.subject):
            item.change_parent(None)
            old_parent.request_update()

        if new_parent and item.subject and group(new_parent.subject, item.subject):
            grow_parent(new_parent, item)
            item.change_parent(new_parent)
        elif item.subject:
            diagram_parent = owner_package(item.diagram)
            group(diagram_parent, item.subject)


@MoveAspect.register(ElementPresentation)
@MoveAspect.register(LinePresentation)
@MoveAspect.register(Presentation)
class DropZoneMove(DropZoneMoveMixin, GuidedItemMoveMixin, ItemMove):
    pass


def grow_parent(parent, item):
    if not isinstance(item, ElementPresentation):
        return

    if not isinstance(parent, ElementPresentation):
        log.warning(f"Can not grow item {parent}: not an ElementPresentation")
        return

    parent_bb = _bounds(parent)
    item_bb = _bounds(item)
    item_bb.expand(20)
    new_parent_bb = parent_bb + item_bb

    c2i = parent.matrix_i2c.inverse()
    parent.handles()[NW].pos = c2i.transform_point(new_parent_bb.x, new_parent_bb.y)
    parent.handles()[SE].pos = c2i.transform_point(new_parent_bb.x1, new_parent_bb.y1)


def _bounds(item):
    transform = item.matrix_i2c.transform_point
    x0, y0 = transform(*item.handles()[NW].pos)
    x1, y1 = transform(*item.handles()[SE].pos)
    return Rectangle(x0, y0, x1=x1, y1=y1)



"""Support code for dealing with stereotypes in diagrams."""

from gaphor.core.format import format
from gaphor.core.styling import JustifyContent
from gaphor.diagram.shapes import Box, CssNode, Text, draw_top_separator


def stereotype_watches(presentation):
    presentation.watch("subject.appliedStereotype", presentation.update_shapes).watch(
        "subject.appliedStereotype.classifier.name"
    ).watch("subject.appliedStereotype.slot", presentation.update_shapes).watch(
        "subject.appliedStereotype.slot.definingFeature.name"
    ).watch("subject.appliedStereotype.slot.value", presentation.update_shapes)


def stereotype_compartments(subject):
    return filter(
        None,
        (
            _create_stereotype_compartment(appliedStereotype)
            for appliedStereotype in subject.appliedStereotype
        )
        if subject
        else [],
    )


def _create_stereotype_compartment(appliedStereotype):
    def lazy_format(slot):
        return lambda: format(slot)

    slots = [slot for slot in appliedStereotype.slot if slot.value]

    if slots:
        return Box(
            CssNode(
                "heading",
                appliedStereotype.classifier,
                Text(
                    text=lazy_format(appliedStereotype.classifier[0])
                    if appliedStereotype.classifier
                    else "",
                ),
            ),
            *(
                CssNode(
                    "slot",
                    slot,
                    Text(
                        text=lazy_format(slot),
                    ),
                )
                for slot in slots
            ),
            style={
                "padding": (4, 4, 4, 4),
                "min-height": 8,
                "justify-content": JustifyContent.START,
            },
            draw=draw_top_separator,
        )
    return None



"""Interface item implementation. There are several notations supported.

- class box with interface stereotype
- folded interface
    - ball is drawn to indicate provided interface
    - socket is drawn to indicate required interface

Interface item can act as icon of assembly connector, see
`gaphor.diagram.connector` module documentation for details. *Documentation
of this module does not take into account assembly connector icon mode.*

Folded Interface Item
=====================
Folded interface notation is reserved for very simple situations.
When interface is folded

- only an implementation can be connected (ball - provided interface)
- or only usage dependency can be connected (socket - required interface)

Above means that interface cannot be folded when

- both, usage dependency and implementation are connected
- any other lines are connected

Dependencies
------------
Dependencies between folded interfaces are *not supported*

+---------------------+---------------------+
|     *Supported*     |    *Unsupported*    |
+=====================+=====================+
| ::                  | ::                  |
|                     |                     |
|   |A|--(    O--|B|  |   |A|--(--->O--|B|  |
|        Z    Z       |        Z    Z       |
+---------------------+---------------------+

On above diagram, A requires interface Z and B provides interface Z.
Additionally, on the right diagram, Z is connected to itself with
dependency.

There is no need for additional dependency

- UML data model provides information, that Z is common for A and B
  (A requires Z, B provides Z)
- on a diagram, both folded interface items (required and provided)
  represent the same interface, which is easily identifiable with its name

Even more, adding a dependency between folded interfaces provides
information, on UML data model level, that an interface depends on itself
but, it is not the intention of this (*unsupported*) notation.

For more examples of non-supported by Gaphor notation, see
http://martinfowler.com/bliki/BallAndSocket.html.


Folding and Connecting
----------------------
Current approach to folding and connecting lines to an interface is as
follows

- allow folding/unfolding of an interface only when there is only one
  implementation or dependency usage connected
- when interface is folded, allow only one implementation or dependency
  usage to be connected

Folding and unfolding is performed by `InterfacePropertyPage` class.
"""

from enum import Enum
from math import pi

from gaphas.connector import LinePort
from gaphas.geometry import distance_line_point, distance_point_point
from gaphas.item import NE, NW, SE, SW

from gaphor import UML
from gaphor.core.modeling.presentation import literal_eval
from gaphor.core.modeling.properties import attribute
from gaphor.core.styling import JustifyContent
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    from_package_str,
    text_name,
)
from gaphor.diagram.shapes import Box, IconBox, Text, draw_border, stroke
from gaphor.diagram.support import represents
from gaphor.UML.classes.klass import (
    attribute_watches,
    attributes_compartment,
    operation_watches,
    operations_compartment,
)
from gaphor.UML.classes.stereotype import stereotype_compartments
from gaphor.UML.recipes import stereotypes_str


class Folded(Enum):
    # Non-folded mode.
    NONE = 0
    # Folded mode, provided (ball) notation.
    PROVIDED = 1
    # Folded mode, required (socket) notation.
    REQUIRED = 2
    # Folded mode, notation of assembly connector icon mode (ball&socket).
    ASSEMBLY = 3


class Side(Enum):
    N = 0
    E = pi * 0.5
    S = pi
    W = pi * 1.5


class InterfacePort(LinePort):
    """Interface connection port.

    It is simple line port, which changes glue behaviour depending on
    interface folded state. If interface is folded, then
    `InterfacePort.glue` method suggests connection in the middle of the
    port.

    The port provides rotation angle information as well. Rotation angle
    is direction the port is facing (i.e. 0 is north, PI/2 is west, etc.).
    The rotation angle shall be used to determine rotation of required
    interface notation (socket's arc is in the same direction as the
    angle).
    """

    def __init__(self, start, end, is_folded, side):
        super().__init__(start, end)
        self.is_folded = is_folded
        # Used by connection logic:
        self.side = side

    def glue(self, pos):
        """Behaves like simple line port, but for folded interface suggests
        connection to the middle point of a port."""
        if self.is_folded():
            px = (self.start.x + self.end.x) / 2
            py = (self.start.y + self.end.y) / 2
            d = distance_point_point((px, py), pos)
            return (px, py), d
        else:
            d, pl = distance_line_point(self.start, self.end, pos)
            return pl, d


@represents(UML.Interface)
class InterfaceItem(Classified, ElementPresentation):
    """Interface item supporting class box, folded notations and assembly
    connector icon mode.

    When in folded mode, provided (ball) notation is used by default.
    """

    RADIUS_PROVIDED = 10
    RADIUS_REQUIRED = 14

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram, id, width=self.RADIUS_PROVIDED * 2, height=self.RADIUS_PROVIDED * 2
        )
        self._folded = Folded.NONE
        self._side = Side.N

        handles = self.handles()
        h_nw = handles[NW]
        h_ne = handles[NE]
        h_sw = handles[SW]
        h_se = handles[SE]

        def is_folded():
            return self._folded != Folded.NONE

        # edge of element define default element ports
        self._ports = [
            InterfacePort(h_nw.pos, h_ne.pos, is_folded, Side.N),
            InterfacePort(h_ne.pos, h_se.pos, is_folded, Side.E),
            InterfacePort(h_se.pos, h_sw.pos, is_folded, Side.S),
            InterfacePort(h_sw.pos, h_nw.pos, is_folded, Side.W),
        ]

        self.watch("show_stereotypes", self.update_shapes).watch(
            "show_attributes", self.update_shapes
        ).watch("show_operations", self.update_shapes).watch(
            "subject[NamedElement].name"
        ).watch("subject[NamedElement].namespace.name").watch(
            "subject.appliedStereotype", self.update_shapes
        ).watch("subject.appliedStereotype.classifier.name").watch(
            "subject.appliedStereotype.slot", self.update_shapes
        ).watch("subject.appliedStereotype.slot.definingFeature.name").watch(
            "subject.appliedStereotype.slot.value", self.update_shapes
        ).watch("subject[Interface].supplierDependency", self.update_shapes)
        attribute_watches(self, "Interface")
        operation_watches(self, "Interface")

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    show_attributes: attribute[int] = attribute("show_attributes", int, default=True)

    show_operations: attribute[int] = attribute("show_operations", int, default=True)

    @property
    def side(self):
        return self._side

    @side.setter
    def side(self, side):
        self._side = side

    def load(self, name, value):
        if name == "folded":
            self._folded = Folded(literal_eval(value))
        else:
            super().load(name, value)

    def save(self, save_func):
        super().save(save_func)
        save_func("folded", self._folded.value)

    def _set_folded(self, folded):
        """Set folded notation.

        :param folded: Folded state, see Folded.* enum.
        """
        if self._folded == folded:
            return

        self._folded = folded

        if folded == Folded.NONE:
            movable = True
        else:
            if self._folded == Folded.PROVIDED:
                icon_size = self.RADIUS_PROVIDED * 2
            else:  # required interface or assembly icon mode
                icon_size = self.RADIUS_REQUIRED * 2

            self.min_width = self.min_height = self.width = self.height = icon_size

            movable = False

        for h in self._handles:
            h.movable = movable

        self.update_shapes()

    folded = property(
        lambda s: s._folded,
        _set_folded,
        doc="Check or set folded notation, see Folded.* enum.",
    )

    def update_shapes(self, event=None):
        connected_items = [
            c.item for c in self.diagram.connections.get_connections(connected=self)
        ]
        connectors = any(isinstance(i.subject, UML.Connector) for i in connected_items)
        if connectors or self._folded != Folded.NONE:
            provided = connectors or any(
                isinstance(i.subject, UML.InterfaceRealization) for i in connected_items
            )
            required = any(isinstance(i.subject, UML.Usage) for i in connected_items)
            if required and provided:
                self.folded = Folded.ASSEMBLY
            elif required:
                self.folded = Folded.REQUIRED
            else:
                self.folded = Folded.PROVIDED

        if self._folded == Folded.NONE:
            self.shape = self.class_shape()
        else:
            self.shape = self.ball_and_socket_shape(connectors)

    def class_shape(self):
        return Box(
            Box(
                Text(
                    text=lambda: stereotypes_str(
                        self.subject, (self.diagram.gettext("interface"),)
                    ),
                ),
                text_name(self),
                Text(
                    text=lambda: from_package_str(self),
                    style={"font-size": "x-small"},
                ),
                style={"padding": (12, 4, 12, 4)},
            ),
            *(
                self.show_attributes
                and self.subject
                and [attributes_compartment(self.subject)]
                or []
            ),
            *(
                self.show_operations
                and self.subject
                and [operations_compartment(self.subject)]
                or []
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            style={
                "justify-content": JustifyContent.START,
            },
            draw=draw_border,
        )

    def ball_and_socket_shape(self, connectors=None):
        if connectors is None:
            # distinguish between None and []
            connected_items = [
                c.item for c in self.diagram.connections.get_connections(connected=self)
            ]
            connectors = any(
                isinstance(i.subject, UML.Connector) for i in connected_items
            )
        return IconBox(
            Box(
                draw=self.draw_interface_ball_and_socket,
            ),
            Text(
                text=lambda: stereotypes_str(self.subject),
            ),
            text_name(self),
        )

    def draw_interface_ball_and_socket(self, _box, context, _bounding_box):
        cr = context.cairo

        h_nw = self._handles[NW]
        cx, cy = (h_nw.pos.x + self.width / 2, h_nw.pos.y + self.height / 2)

        if self._folded in (Folded.REQUIRED, Folded.ASSEMBLY):
            r = self.RADIUS_REQUIRED
            if self._side == Side.N:
                x, y = r * 2, r
            elif self._side == Side.E:
                x, y = r, r * 2
            elif self._side == Side.S:
                x, y = 0, r
            elif self._side == Side.W:
                x, y = r, 0

            cr.move_to(x, y)
            cr.arc_negative(
                cx, cy, self.RADIUS_REQUIRED, self._side.value, pi + self._side.value
            )

        if self._folded in (Folded.PROVIDED, Folded.ASSEMBLY):
            cr.move_to(cx + self.RADIUS_PROVIDED, cy)
            cr.arc(cx, cy, self.RADIUS_PROVIDED, 0, pi * 2)

        stroke(context, fill=True)



from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.diagram.presentation import (
    AttachedPresentation,
    Classified,
    ElementPresentation,
    connect,
)
from gaphor.diagram.shapes import Box, JustifyContent, Text, TextAlign, draw_border
from gaphor.diagram.support import represents
from gaphor.diagram.text import FontStyle
from gaphor.UML.recipes import stereotypes_str


@represents(UML.Activity)
class ActivityItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=50, height=50)

        self.width = 100

        self.watch("subject[NamedElement].name").watch(
            "subject.appliedStereotype.classifier.name"
        ).watch("subject[Classifier].isAbstract", self.update_shapes).watch(
            "subject[Activity].node[ActivityParameterNode].parameter.name",
            self.update_parameters,
        ).watch(
            "subject[Activity].node[ActivityParameterNode].parameter.typeValue",
            self.update_parameters,
        )

    def postload(self):
        super().postload()
        self.update_parameters()

    def update_shapes(self, event=None):
        self.shape = Box(
            Text(
                text=lambda: stereotypes_str(self.subject),
                style={"text-align": TextAlign.LEFT},
            ),
            Text(
                text=lambda: self.subject.name or "",
                style={
                    "text-align": TextAlign.LEFT,
                    "font-style": FontStyle.ITALIC
                    if self.subject and self.subject.isAbstract
                    else FontStyle.NORMAL,
                },
            ),
            style={
                "padding": (4, 12, 4, 12),
                "border-radius": 20,
                "justify-content": JustifyContent.START,
            },
            draw=draw_border,
        )

    def update_parameters(self, event=None):
        diagram = self.diagram
        parameter_nodes = (
            [p for p in self.subject.node if isinstance(p, UML.ActivityParameterNode)]
            if self.subject
            else []
        )
        parameter_items = {
            i.subject: i
            for i in self.children
            if isinstance(i, ActivityParameterNodeItem)
        }

        for node in parameter_nodes:
            if node not in parameter_items:
                item = diagram.create(
                    ActivityParameterNodeItem, parent=self, subject=node
                )
                item.matrix.translate(0, 10)
                connect(item, item.handles()[0], self)

        for node in parameter_items:
            if node not in parameter_nodes:
                del self.children[parameter_items[node]]


@represents(UML.ActivityParameterNode)
class ActivityParameterNodeItem(AttachedPresentation[UML.ActivityParameterNode]):
    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape=Box(
                Text(
                    text=self._format_name,
                ),
                style={"padding": (4, 12, 4, 12), "min-width": 120},
                draw=draw_border,
            ),
        )

        self.watch("subject[ActivityParameterNode].parameter.name").watch(
            "subject[ActivityParameterNode].parameter.type.name"
        ).watch("subject[ActivityParameterNode].parameter.direction").watch(
            "show_type"
        ).watch("show_direction")

    show_type: attribute[int] = attribute("show_type", int, default=False)
    show_direction: attribute[int] = attribute("show_direction", int, default=False)

    def update(self, context):
        self.width, self.height = super().update(context)

    def _format_name(self):
        if not (self.subject and self.subject.parameter):
            return ""
        parameter = self.subject.parameter
        name = parameter.name or ""
        type = (parameter.type.name) if self.show_type and parameter.type else None
        direction = parameter.direction if self.show_direction else None

        if type and direction:
            return f"{direction} {name}: {type}"
        elif type:
            return f"{name}: {type}"
        elif direction:
            return f"{direction} {name}"
        return name



import io

import defusedxml.ElementTree as etree

from gaphor.i18n import gettext


def extract_gaphor(fileobj, keywords, comment_tags, options):
    """Extract text from Gaphor models.

    Args:
        fileobj (obj): the file-like object the messages should be extracted from.
        keywords (list): a list of keywords (i.e. function names) that should
              be recognized as translation functions.
        comment_tags (list): a list of translator tags to search for and
                  include in the results.
        options (dict, optional) : a dictionary of additional options.
    Returns:
        iterator: an iterator over ``(lineno, funcname, message, comments)``
             tuples.

    See also:

    * babel.messages.extract.extract()
    * http://babel.pocoo.org/en/latest/messages.html#writing-extraction-methods
    * https://www.gnu.org/software/gettext/manual/html_node/PO-Files.html
    """
    lineno = None
    funcname = "gettext"
    comments: list[str] = []

    tree = etree.parse(fileobj)

    for node in translatable_nodes(tree):
        yield (lineno, funcname, node.text, comments)


def translate_model(fileobj):
    tree = etree.parse(fileobj)

    for node in translatable_nodes(tree):
        node.text = gettext(node.text) if node.text else ""

    return io.StringIO(etree.tostring(tree.getroot(), encoding="unicode", method="xml"))


def translatable_nodes(tree):
    NS = {"g": "http://gaphor.sourceforge.net/model"}

    for tag in ("name", "body", "typeValue"):
        yield from tree.findall(f".//g:{tag}/g:val", NS)



from gaphor.core.styling import CompiledStyleSheet, Style, StyleNode


def compute_inherited_style(style: Style, style_node: StyleNode) -> Style:
    compiled_style_sheet: CompiledStyleSheet | None = style.get(
        "-gaphor-compiled-style-sheet"
    )  # type: ignore[assignment]

    return (
        compiled_style_sheet.compute_style(style_node)
        if compiled_style_sheet
        else style
    )



"""The file service is responsible for loading and saving the user data."""

from __future__ import annotations

import logging
import tempfile
from functools import partial
from pathlib import Path
from typing import Callable

from gaphas.decorators import g_async
from gi.repository import Adw, Gtk

from gaphor import UML
from gaphor.abc import ActionProvider, Service
from gaphor.core import action, event_handler, gettext
from gaphor.core.changeset.compare import compare
from gaphor.core.modeling import Diagram, ElementFactory, StyleSheet
from gaphor.event import (
    ModelLoaded,
    ModelSaved,
    SessionCreated,
    SessionShutdown,
    SessionShutdownRequested,
)
from gaphor.storage import storage
from gaphor.storage.mergeconflict import split_ours_and_theirs
from gaphor.storage.parser import MergeConflictDetected
from gaphor.ui.errorhandler import error_handler
from gaphor.ui.filedialog import GAPHOR_FILTER, save_file_dialog
from gaphor.ui.statuswindow import StatusWindow

DEFAULT_EXT = ".gaphor"
MAX_RECENT = 10

log = logging.getLogger(__name__)


def error_message(e):
    if not isinstance(e, IOError):
        return gettext(
            "Gaphor was not able to store the model, probably due to an internal error:\n{exc}\nIf you think this is a bug, please contact the developers."
        ).format(exc=str(e))
    if e.errno == 13:
        return gettext(
            "You do not have the permissions necessary to save the model.\nPlease check that you typed the location correctly and try again."
        )
    elif e.errno == 28:
        return gettext(
            "You do not have enough free space on the device to save the model.\nPlease free up some disk space and try again or save it in a different location."
        )
    return gettext(
        "The model cannot be stored at this location:\n{exc}\nPlease check that you typed the location correctly and try again."
    ).format(exc=str(e))


def load_default_model(element_factory):
    element_factory.flush()
    with element_factory.block_events():
        element_factory.create(StyleSheet)
        model = element_factory.create(UML.Package)
        model.name = gettext("New model")
        diagram = element_factory.create(Diagram)
        diagram.element = model
        diagram.name = gettext("New diagram")
    element_factory.model_ready()


class FileManager(Service, ActionProvider):
    """The file service, responsible for loading and saving Gaphor models."""

    def __init__(self, event_manager, element_factory, modeling_language, main_window):
        """File manager constructor.

        There is no current filename yet.
        """
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.modeling_language = modeling_language
        self.main_window = main_window
        self._filename: Path | None = None

        event_manager.subscribe(self._on_session_shutdown_request)
        event_manager.subscribe(self._on_session_created)

    def shutdown(self):
        """Called when shutting down the file manager service."""
        self.event_manager.unsubscribe(self._on_session_shutdown_request)
        self.event_manager.unsubscribe(self._on_session_created)

    @property
    def filename(self) -> Path | None:
        """Return the current file name.

        This method is used by the filename property.
        """
        return self._filename

    @filename.setter
    def filename(self, filename: Path | str | None):
        """Sets the current file name.

        This method is used by the filename property.
        """

        if filename != self._filename:
            self._filename = Path(filename) if filename else None

    def load_template(self, template):
        storage.load(template, self.element_factory, self.modeling_language)
        self.event_manager.handle(ModelLoaded(self))

    def load(self, filename: Path, on_load_done: Callable[[], None] | None = None):
        """Load the Gaphor model from the supplied file name.

        A status window displays the loading progress. The load
        generator updates the progress queue.  The loader is passed to a
        GIdleThread which executes the load generator. If loading is
        successful, the filename is set.
        """
        # First claim file name, so any other files will be opened in a different session
        self.filename = filename

        status_window = StatusWindow(
            gettext("Loading…"),
            gettext("Loading model from {filename}").format(filename=filename),
            parent=self.parent_window,
        )

        def done():
            status_window.destroy()
            if on_load_done:
                on_load_done()
            else:
                self.event_manager.handle(ModelLoaded(self, filename))

        for _ in self._load_async(filename, status_window.progress, done):
            pass

    def merge(
        self,
        ancestor_filename: Path,
        current_filename: Path,
        incoming_filename: Path,
        on_load_done: Callable[[], None] | None = None,
    ):
        status_window = StatusWindow(
            gettext("Loading…"),
            gettext("Loading current and incoming model"),
            parent=self.parent_window,
        )

        ancestor_element_factory = ElementFactory()
        incoming_element_factory = ElementFactory()

        def progress(percentage, completed=0):
            status_window.progress(completed + percentage / 3)

        # Make this callback async, so we can call _load_async again: it's a generator and we can only run one at a time
        @g_async()
        def current_done():
            log.debug("Loading ancestor model from %s", ancestor_filename)
            for _ in self._load_async(
                ancestor_filename,
                partial(progress, completed=33),
                ancestor_done,
                element_factory=ancestor_element_factory,
            ):
                pass

        @g_async()
        def ancestor_done():
            log.debug("Loading incoming model from %s", incoming_filename)
            for _ in self._load_async(
                incoming_filename,
                partial(progress, completed=66),
                incoming_done,
                element_factory=incoming_element_factory,
            ):
                pass

        def incoming_done():
            try:
                log.debug("Comparing models")
                with self.element_factory.block_events():
                    list(
                        compare(
                            self.element_factory,
                            ancestor_element_factory,
                            incoming_element_factory,
                        )
                    )

                if on_load_done:
                    on_load_done()
            finally:
                status_window.destroy()

        log.debug("Loading current model from %s", current_filename)
        for _ in self._load_async(current_filename, progress, current_done):
            pass

    @g_async()
    def _load_async(
        self,
        filename: Path,
        progress: Callable[[int], None] | None = None,
        done=None,
        element_factory=None,
    ):
        try:
            with filename.open(encoding="utf-8", errors="replace") as file_obj:
                for percentage in storage.load_generator(
                    file_obj,
                    element_factory or self.element_factory,
                    self.modeling_language,
                ):
                    if progress:
                        progress(percentage)
                    yield percentage
        except MergeConflictDetected:
            self.filename = None
            self.resolve_merge_conflict(filename)
        except Exception:
            self.filename = None
            error_handler(
                message=gettext("Unable to open model “{filename}”.").format(
                    filename=filename.name
                ),
                secondary_message=gettext(
                    "This file does not contain a valid Gaphor model."
                ),
                window=self.parent_window,
                close=lambda: self.event_manager.handle(SessionShutdown(self)),
            )
        finally:
            if done:
                done()

    def resolve_merge_conflict(self, filename: Path):
        temp_dir = tempfile.TemporaryDirectory()
        ancestor_filename = Path(temp_dir.name) / f"ancestor-{filename.name}"
        current_filename = Path(temp_dir.name) / f"current-{filename.name}"
        incoming_filename = Path(temp_dir.name) / f"incoming-{filename.name}"
        with (
            ancestor_filename.open("wb") as ancestor_file,
            current_filename.open("wb") as current_file,
            incoming_filename.open("wb") as incoming_file,
        ):
            split = split_ours_and_theirs(
                filename, ancestor_file, current_file, incoming_file
            )

        def done():
            nonlocal temp_dir
            temp_dir.cleanup()
            self.filename = filename
            self.event_manager.handle(ModelLoaded(self, filename, modified=True))

        def handle_merge_conflict(answer):
            if answer == "cancel":
                self.event_manager.handle(SessionShutdown(self))
            elif answer == "current":
                self.load(current_filename, on_load_done=done)
            elif answer == "incoming":
                self.load(incoming_filename, on_load_done=done)
            elif answer == "manual":
                self.merge(
                    ancestor_filename,
                    current_filename,
                    incoming_filename,
                    on_load_done=done,
                )
            else:
                raise ValueError(f"Unknown resolution for merge conflict: {answer}")

        if split:
            resolve_merge_conflict_dialog(
                self.parent_window, filename, handle_merge_conflict
            )
        else:
            error_handler(
                message=gettext("Unable to open model “{filename}”.").format(
                    filename=filename.name
                ),
                secondary_message=gettext(
                    "This file does not contain a valid Gaphor model."
                ),
                window=self.parent_window,
                close=lambda: self.event_manager.handle(SessionShutdown(self)),
            )

    def save(self, filename, on_save_done=None):
        """Save the current UML model to the specified file name.

        Before writing the model file, this will verify that there are
        no orphan references.  It will also verify that the filename has
        the correct extension.  A status window is displayed while the
        GIdleThread is executed.  This thread actually saves the model.
        """

        if not filename or (filename.exists() and not filename.is_file()):
            return

        status_window = StatusWindow(
            gettext("Saving…"),
            gettext("Saving model to {filename}").format(filename=filename),
            parent=self.parent_window,
        )

        @g_async()
        def async_saver():
            try:
                with filename.open("w", encoding="utf-8") as out:
                    for percentage in storage.save_generator(out, self.element_factory):
                        status_window.progress(percentage)
                        yield

                self.filename = filename
                self.event_manager.handle(ModelSaved(self, filename))
            except Exception as e:
                error_handler(
                    message=gettext("Unable to save model “{filename}”.").format(
                        filename=filename
                    ),
                    secondary_message=error_message(e),
                    window=self.parent_window,
                )
                raise
            finally:
                status_window.destroy()
            if on_save_done:
                on_save_done()

        for _ in async_saver():
            pass

    @property
    def parent_window(self):
        return self.main_window.window if self.main_window else None

    @action(name="file-save", shortcut="<Primary>s")
    def action_save(self):
        """Save the file. Depending on if there is a file name, either perform
        the save directly or present the user with a save dialog box.

        Returns True if the saving actually succeeded.
        """

        if filename := self.filename:
            self.save(filename)
        else:
            self.action_save_as()

    @action(name="file-save-as", shortcut="<Primary><Shift>s")
    def action_save_as(self):
        """Save the model in the element_factory by allowing the user to select
        a file name."""

        return save_file_dialog(
            gettext("Save Gaphor Model As"),
            self.save,
            parent=self.parent_window,
            filename=self.filename,
            extension=".gaphor",
            filters=GAPHOR_FILTER,
        )

    @event_handler(SessionCreated)
    def _on_session_created(self, event: SessionCreated) -> None:
        if event.filename:
            self.load(event.filename)
        elif event.template:
            self.load_template(event.template)
        else:
            load_default_model(self.element_factory)

    @event_handler(SessionShutdownRequested)
    def _on_session_shutdown_request(self, event: SessionShutdownRequested) -> None:
        """Ask user to close window if the model has changed.

        The user is asked to either discard the changes, keep the
        application running or save the model and quit afterwards.
        """

        def confirm_shutdown():
            self.event_manager.handle(SessionShutdown(self))

        def response(answer):
            if answer == "save":
                if filename := self.filename:
                    self.save(filename, on_save_done=confirm_shutdown)

                else:
                    save_file_dialog(
                        gettext("Save Gaphor Model As"),
                        lambda filename: self.save(
                            filename, on_save_done=confirm_shutdown
                        ),
                        parent=self.parent_window,
                        filename=self.filename,
                        extension=".gaphor",
                        filters=GAPHOR_FILTER,
                    )
            elif answer == "discard":
                confirm_shutdown()

        if self.main_window.model_changed:
            save_changes_before_close_dialog(self.parent_window, response)
        else:
            confirm_shutdown()


def resolve_merge_conflict_dialog(window: Gtk.Window, filename: Path, handler) -> None:
    dialog = Adw.MessageDialog.new(
        window,
        gettext("Resolve Merge Conflict?"),
    )
    dialog.set_body(
        gettext(
            "The model you are opening contains a merge conflict. Do you want to open the current model or the incoming change to the model?"
        )
    )
    dialog.add_response("cancel", gettext("Cancel"))
    dialog.add_response("manual", gettext("Open Merge Editor"))
    dialog.add_response("current", gettext("Open Current"))
    dialog.add_response("incoming", gettext("Open Incoming"))
    dialog.set_close_response("cancel")

    def response(dialog, answer):
        dialog.set_transient_for(None)
        dialog.destroy()
        handler(answer)

    dialog.connect("response", response)
    dialog.set_visible(True)


def save_changes_before_close_dialog(window: Gtk.Window, handler) -> None:
    title = gettext("Save Changes?")
    body = gettext(
        "The open model contains unsaved changes. Changes which are not saved will be permanently lost."
    )
    dialog = Adw.MessageDialog.new(
        window,
        title,
    )
    dialog.set_body(body)
    dialog.add_response("cancel", gettext("Cancel"))
    dialog.add_response("discard", gettext("Discard"))
    dialog.add_response("save", gettext("Save"))
    dialog.set_response_appearance("discard", Adw.ResponseAppearance.DESTRUCTIVE)
    dialog.set_response_appearance("save", Adw.ResponseAppearance.SUGGESTED)
    dialog.set_default_response("save")
    dialog.set_close_response("cancel")

    def response(dialog, answer):
        # Unset transient window: it can cause crashes on flatpak
        # when all windows are destroyed at once.
        dialog.set_transient_for(None)
        dialog.destroy()
        handler(answer)

    dialog.connect("response", response)
    dialog.set_visible(True)



"""Factory for and registration of model elements."""

from __future__ import annotations

from collections import OrderedDict
from contextlib import contextmanager
from typing import Callable, Iterator, Protocol, TypeVar, overload

from gaphor.abc import Service
from gaphor.core.eventmanager import EventManager, event_handler
from gaphor.core.modeling.diagram import Diagram
from gaphor.core.modeling.element import (
    Element,
    EventWatcherProtocol,
    Handler,
    Id,
    RepositoryProtocol,
    UnlinkEvent,
    generate_id,
)
from gaphor.core.modeling.elementdispatcher import ElementDispatcher, EventWatcher
from gaphor.core.modeling.event import (
    ElementCreated,
    ElementDeleted,
    ModelFlushed,
    ModelReady,
)
from gaphor.core.modeling.presentation import Presentation

T = TypeVar("T", bound=Element)
P = TypeVar("P", bound=Presentation)


class EventHandler(Protocol):
    def handle(self, *events):
        ...


class RecordingEventManager:
    def __init__(self, event_manager):
        self.event_manager = event_manager
        self.events = []

    def handle(self, *events):
        self.events.extend(events)

    def replay(self):
        if self.event_manager:
            self.event_manager.handle(*self.events)


class ElementFactory(Service):
    """The ElementFactory is used to create elements and do lookups to
    elements.

    Notifications are sent as arguments (name, element, `*user_data`).
    The following names are used:
    create - a new model element is created (element is newly created element)
    remove - a model element is removed (element is to be removed element)
    model - a new model has been loaded (element is None) flush - model is
      flushed: all element are removed from the factory (element is None)
    """

    def __init__(
        self,
        event_manager: EventManager | None = None,
        element_dispatcher: ElementDispatcher | None = None,
    ):
        self.event_manager: EventHandler | None = event_manager
        self.element_dispatcher = element_dispatcher
        self._elements: dict[Id, Element] = OrderedDict()
        if event_manager:
            event_manager.subscribe(self._on_unlink_event)

    def shutdown(self) -> None:
        self.flush()
        if isinstance(self.event_manager, EventManager):
            self.event_manager.unsubscribe(self._on_unlink_event)

    def create(self, type: type[T]) -> T:
        """Create a new model element of type ``type``."""
        return self.create_as(type, generate_id())

    def create_as(self, type: type[T], id: Id, diagram: Diagram | None = None) -> T:
        """Create a new model element of type 'type' with 'id' as its ID.

        This method should only be used when loading models, since it
        does not emit an ElementCreated event.
        """
        if id in self._elements:
            element = self._elements[id]
            if not isinstance(element, type):
                raise TypeError(
                    "Element {element} already exists but has a different type  {type}"
                )
            return element

        type_args: dict[str, Diagram | RepositoryProtocol]
        if issubclass(type, Presentation):
            if not diagram:
                raise TypeError("Presentation types require a diagram")
            type_args = {"diagram": diagram}
        elif issubclass(type, Element):
            if diagram:
                raise TypeError("Element types require no diagram")
            type_args = {"model": self}
        else:
            raise TypeError(f"Type {type} is not a valid model element")

        # Avoid events that reference this element before its created-event is emitted.
        event_recorder = RecordingEventManager(self.event_manager)
        with self.block_events(event_recorder):
            element = type(id=id, **type_args)  # type: ignore[arg-type]
        self._elements[id] = element
        self.handle(ElementCreated(self, element, diagram))
        event_recorder.replay()
        return element

    def size(self) -> int:
        """Return the amount of elements currently in the factory."""
        return len(self._elements)

    def lookup(self, id: Id) -> Element | None:
        """Find element with a specific id."""
        return self._elements.get(id)

    def __getitem__(self, id: Id) -> Element:
        return self._elements[id]

    def __iter__(self):
        return iter(self._elements.values())

    def __contains__(self, element: Element) -> bool:
        assert isinstance(element.id, Id)
        return self.lookup(element.id) is element

    @overload
    def select(self, expression: Callable[[Element], bool]) -> Iterator[Element]:
        ...

    @overload
    def select(self, expression: type[T]) -> Iterator[T]:
        ...

    @overload
    def select(self, expression: None) -> Iterator[Element]:
        ...

    def select(self, expression=None):
        """Iterate elements that comply with expression."""
        if expression is None:
            yield from self._elements.values()
        elif isinstance(expression, type):
            yield from (e for e in self._elements.values() if isinstance(e, expression))
        else:
            yield from (e for e in self._elements.values() if expression(e))

    def lselect(
        self, expression: Callable[[Element], bool] | type[T] | None = None
    ) -> list[Element]:
        """Get a list of elements that comply with expression.

        The expression can be one of:
        - None: return all elements
        - class name: return all elements of type `class name`
        - expression(e: Element) -> bool: return elements that comply with expression
        """
        return list(self.select(expression))

    def keys(self) -> Iterator[Id]:
        """Return a list with all id's in the factory."""
        return iter(self._elements.keys())

    def values(self) -> Iterator[Element]:
        """Return a list with all elements in the factory."""
        return iter(self._elements.values())

    def is_empty(self) -> bool:
        """Returns True if the factory holds no elements."""
        return bool(self._elements)

    def watcher(
        self, element: Element, default_handler: Handler | None = None
    ) -> EventWatcherProtocol:
        element_dispatcher = self.element_dispatcher
        return EventWatcher(element, element_dispatcher, default_handler)

    def flush(self) -> None:
        """Flush all elements (remove them from the factory).

        Diagram elements are flushed first. The remaining elements are
        flushed next.
        """
        with self.block_events():
            for element in self.lselect(Diagram):
                assert isinstance(element, Diagram)
                element.unlink()

            for element in self.lselect():
                element.unlink()

        self.handle(ModelFlushed(self))

    def model_ready(self) -> None:
        """Send notification that a new model has been loaded by means of the
        ModelReady event from gaphor.core.modeling.event."""
        self.handle(ModelReady(self))

    @contextmanager
    def block_events(self, new_event_manager: EventHandler | None = None):
        """Block events from being emitted.

        Instead, events are directed to `new_event_manager`, which
        defaults to a blocking event manager.
        """
        current_event_manager = self.event_manager
        self.event_manager = new_event_manager
        try:
            yield self
        finally:
            self.event_manager = current_event_manager

    def handle(self, event: object) -> None:
        """Handle events coming from elements."""
        if self.event_manager:
            self.event_manager.handle(event)
        elif isinstance(event, UnlinkEvent):
            self._on_unlink_event(event)

    @event_handler(UnlinkEvent)
    def _on_unlink_event(self, event):
        element = event.element
        element._model = None
        assert isinstance(element.id, Id)
        try:
            del self._elements[element.id]
        except KeyError:
            return
        if self.event_manager:
            self.event_manager.handle(
                ElementDeleted(self, event.element, event.diagram)
            )



import logging

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.core.styling import JustifyContent
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    from_package_str,
    text_name,
)
from gaphor.diagram.shapes import Box, CssNode, Text, draw_border, draw_top_separator
from gaphor.diagram.support import represents
from gaphor.UML.classes.klass import (
    attribute_watches,
    attributes_compartment,
    operation_watches,
    operations_compartment,
)
from gaphor.UML.classes.stereotype import stereotype_compartments, stereotype_watches
from gaphor.UML.recipes import stereotypes_str

log = logging.getLogger(__name__)


@represents(UML.Enumeration)
class EnumerationItem(Classified, ElementPresentation[UML.Enumeration]):
    """This item visualizes an Enumeration instance.

    An EnumerationItem contains three compartments:
    1. Enumeration Literals
    2. Attributes
    3. Operations
    """

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id=id)

        self.watch("show_attributes", self.update_shapes).watch(
            "show_operations", self.update_shapes
        ).watch("show_enumerations", self.update_shapes).watch(
            "subject[NamedElement].name"
        ).watch("subject[NamedElement].namespace.name").watch(
            "subject[Enumeration].ownedLiteral", self.update_shapes
        ).watch("subject[Enumeration].ownedLiteral.name", self.update_shapes)
        attribute_watches(self, "Enumeration")
        operation_watches(self, "Enumeration")
        stereotype_watches(self)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    show_attributes: attribute[int] = attribute("show_attributes", int, default=True)

    show_operations: attribute[int] = attribute("show_operations", int, default=True)

    show_enumerations: attribute[int] = attribute(
        "show_enumerations", int, default=True
    )

    as_sysml_value_type: attribute[int] = attribute(
        "as_sysml_value_type", int, default=False
    )

    def update_shapes(self, event=None):
        self.shape = Box(
            Box(
                Text(
                    text=lambda: stereotypes_str(
                        self.subject,
                        [
                            self.diagram.gettext("valueType")
                            if self.as_sysml_value_type
                            else self.diagram.gettext("enumeration")
                        ],
                    ),
                ),
                text_name(self),
                Text(
                    text=lambda: from_package_str(self),
                    style={"font-size": "x-small"},
                ),
                style={"padding": (12, 4, 12, 4)},
            ),
            *(
                self.show_enumerations
                and self.subject
                and [enumerations_compartment(self.subject)]
                or []
            ),
            *(
                self.show_attributes
                and self.subject
                and [attributes_compartment(self.subject)]
                or []
            ),
            *(
                self.show_operations
                and self.subject
                and [operations_compartment(self.subject)]
                or []
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            style={
                "justify-content": JustifyContent.START,
            },
            draw=draw_border,
        )


def enumerations_compartment(subject):
    def lazy_format(literal):
        return lambda: format(literal)

    return Box(
        *(
            CssNode(
                "enumeration",
                literal,
                Text(
                    text=lazy_format(literal.name),
                ),
            )
            for literal in subject.ownedLiteral
        ),
        style={
            "padding": (4, 4, 4, 4),
            "min-height": 8,
            "justify-content": JustifyContent.START,
        },
        draw=draw_top_separator,
    )



from gaphor import UML
from gaphor.diagram.connectors import Connector, DirectionalRelationshipConnect
from gaphor.diagram.presentation import Classified
from gaphor.UML.profiles.extension import ExtensionItem
from gaphor.UML.recipes import owner_package


@Connector.register(Classified, ExtensionItem)
class ExtensionConnect(DirectionalRelationshipConnect):
    """Connect class and stereotype items using an extension item."""

    line: ExtensionItem

    def allow(self, handle, port):
        line = self.line
        subject = self.element.subject

        if handle is line.head:
            # Element at the head should be a class
            # (implies stereotype as well)
            allow = isinstance(subject, UML.Class)
        elif handle is line.tail:
            # Element at the tail should be a stereotype
            allow = isinstance(subject, UML.Stereotype)

        return allow and super().allow(handle, port)

    def connect_subject(self, handle):
        element = self.element
        line = self.line
        assert element.diagram

        c1 = self.get_connected(line.head)
        c2 = self.get_connected(line.tail)
        if c1 and c2:
            assert isinstance(c1.subject, UML.Class)
            assert isinstance(c2.subject, UML.Stereotype)

            head_type: UML.Class = c1.subject
            tail_type: UML.Stereotype = c2.subject

            # First check if we do not already contain the right subject:
            if line.subject:
                end1 = line.subject.memberEnd[0]
                end2 = line.subject.memberEnd[1]
                if (end1.type is head_type and end2.type is tail_type) or (
                    end2.type is head_type and end1.type is tail_type
                ):
                    return

            c1.request_update()

            # Find all associations and determine if the properties on
            # the association ends have a type that points to the class.
            ext: UML.Extension
            for ext in line.model.select(UML.Extension):
                end1 = ext.memberEnd[0]
                end2 = ext.memberEnd[1]
                if (end1.type is head_type and end2.type is tail_type) or (
                    end2.type is head_type and end1.type is tail_type
                ):
                    # check if this entry is not yet in the diagram
                    # Return if the association is not (yet) on the diagram
                    for item in ext.presentation:
                        if item.diagram is element.diagram:
                            break
                    else:
                        line.subject = ext
                        return
            # Create a new Extension relationship
            relation = UML.recipes.create_extension(head_type, tail_type)
            relation.package = owner_package(element.diagram.owner)
            line.subject = relation



from gaphor.core.modeling.properties import attribute
from gaphor.core.styling import (
    FontWeight,
    JustifyContent,
    Style,
    TextAlign,
    VerticalAlign,
)
from gaphor.diagram.presentation import ElementPresentation
from gaphor.diagram.shapes import (
    Box,
    Orientation,
    Text,
    draw_border,
    draw_left_separator,
    draw_top_separator,
)


class MetadataItem(ElementPresentation):
    createdBy: attribute[str] = attribute("createdBy", str, "")
    description: attribute[str] = attribute("description", str, "")
    website: attribute[str] = attribute("website", str, "")
    revision: attribute[str] = attribute("revision", str, "")
    license: attribute[str] = attribute("license", str, "")
    createdOn: attribute[str] = attribute("createdOn", str, "")
    updatedOn: attribute[str] = attribute("updatedOn", str, "")

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("createdBy", self.update_shapes).watch(
            "website", self.update_shapes
        ).watch("description", self.update_shapes).watch(
            "revision", self.update_shapes
        ).watch("license", self.update_shapes).watch(
            "createdOn", self.update_shapes
        ).watch("updatedOn", self.update_shapes)

    def update_shapes(self, event=None):
        diagram = self.diagram
        group_style: Style = {"justify-content": JustifyContent.STRETCH}
        box_style: Style = {"padding": (4, 4, 4, 4)}
        text_style: Style = {
            "text-align": TextAlign.LEFT,
            "vertical-align": VerticalAlign.TOP,
        }
        heading_style: Style = {
            **text_style,
            "font-size": "small",
            "font-weight": FontWeight.BOLD,
        }

        self.shape = Box(
            *(
                [
                    Box(
                        *(
                            [
                                Box(
                                    Text(
                                        text=f'{diagram.gettext("Created By")}:',
                                        style=heading_style,
                                    ),
                                    Text(
                                        text=lambda: self.createdBy or "",
                                        style=text_style,
                                    ),
                                    style=box_style,
                                )
                            ]
                            if self.createdBy
                            else []
                        ),
                        *(
                            [
                                Box(
                                    Text(
                                        text=f'{diagram.gettext("Website")}:',
                                        style=heading_style,
                                    ),
                                    Text(
                                        text=lambda: self.website or "",
                                        style=text_style,
                                    ),
                                    style=box_style,
                                    draw=draw_left_separator,
                                )
                            ]
                            if self.website
                            else []
                        ),
                        orientation=Orientation.HORIZONTAL,
                        style=group_style,
                    )
                ]
                if self.createdBy or self.website
                else []
            ),
            *(
                [
                    Box(
                        Text(
                            text=f'{diagram.gettext("Description")}:',
                            style=heading_style,
                        ),
                        Text(
                            text=lambda: self.description or "",
                            style=text_style,
                        ),
                        style=box_style,
                        draw=draw_top_separator,
                    )
                ]
                if self.description
                else []
            ),
            *(
                [
                    Box(
                        *(
                            [
                                Box(
                                    Text(
                                        text=f'{diagram.gettext("Revision")}:',
                                        style=heading_style,
                                    ),
                                    Text(
                                        text=lambda: self.revision or "",
                                        style=text_style,
                                    ),
                                    style=box_style,
                                )
                            ]
                            if self.revision
                            else []
                        ),
                        *(
                            [
                                Box(
                                    Text(
                                        text=f'{diagram.gettext("License")}:',
                                        style=heading_style,
                                    ),
                                    Text(
                                        text=lambda: self.license or "",
                                        style=text_style,
                                    ),
                                    style=box_style,
                                    draw=draw_left_separator,
                                )
                            ]
                            if self.license
                            else []
                        ),
                        *(
                            [
                                Box(
                                    Text(
                                        text=f'{diagram.gettext("Created On")}:',
                                        style=heading_style,
                                    ),
                                    Text(
                                        text=lambda: self.createdOn or "",
                                        style=text_style,
                                    ),
                                    style=box_style,
                                    draw=draw_left_separator,
                                )
                            ]
                            if self.createdOn
                            else []
                        ),
                        *(
                            [
                                Box(
                                    Text(
                                        text=f'{diagram.gettext("Updated On")}:',
                                        style=heading_style,
                                    ),
                                    Text(
                                        text=lambda: self.updatedOn or "",
                                        style=text_style,
                                    ),
                                    style=box_style,
                                    draw=draw_left_separator,
                                )
                            ]
                            if self.updatedOn
                            else []
                        ),
                        orientation=Orientation.HORIZONTAL,
                        style=group_style,
                        draw=draw_top_separator,
                    )
                ]
                if self.revision or self.license or self.createdOn or self.updatedOn
                else []
            ),
            draw=draw_border,
            style=group_style,
        )



#!/usr/bin/python

import argparse
import logging
import re
from pathlib import Path
from typing import List

from gaphor.application import Session
from gaphor.core.modeling import Diagram
from gaphor.diagram.export import escape_filename, save_pdf, save_png, save_svg
from gaphor.storage import storage

log = logging.getLogger(__name__)


def pkg2dir(package):
    """Return directory path from package class."""
    name: List[str] = []
    while package:
        name.insert(0, package.name)
        package = package.package
    return "/".join(name)


def export_parser():
    parser = argparse.ArgumentParser(description="Export diagrams from a Gaphor model.")

    parser.add_argument(
        "-u",
        "--use-underscores",
        dest="underscores",
        action="store_true",
        help="use underscores instead of spaces for output filenames",
    )
    parser.add_argument(
        "-o", "--dir", metavar="directory", help="output to directory", default="."
    )
    parser.add_argument(
        "-f",
        "--format",
        metavar="format",
        help="output file format, default pdf",
        default="pdf",
        choices=["pdf", "svg", "png"],
    )
    parser.add_argument(
        "-r",
        "--regex",
        dest="regex",
        metavar="regex",
        help="process diagrams which name matches given regular expression;"
        " name includes package name; regular expressions are case insensitive",
    )
    parser.add_argument("model", nargs="+")
    parser.set_defaults(command=export_command)

    return parser


def export_command(args):
    session = Session(
        services=[
            "event_manager",
            "component_registry",
            "element_factory",
            "element_dispatcher",
            "modeling_language",
        ]
    )
    factory = session.get_service("element_factory")
    modeling_language = session.get_service("modeling_language")

    name_re = re.compile(args.regex, re.IGNORECASE) if args.regex else None
    # we should have some gaphor files to be processed at this point
    for model in args.model:
        log.debug("loading model %s", model)
        with open(model, encoding="utf-8") as file_obj:
            storage.load(file_obj, factory, modeling_language)
        log.debug("ready for rendering")

        for diagram in factory.select(Diagram):
            odir = pkg2dir(diagram.owner)

            # just diagram name
            dname = escape_filename(diagram.name)
            # full diagram name including package path
            pname = f"{odir}/{dname}"

            if args.underscores:
                odir = odir.replace(" ", "_")
                dname = dname.replace(" ", "_")

            if name_re and not name_re.search(pname):
                log.debug("skipping %s", pname)
                continue

            if args.dir:
                odir = f"{args.dir}/{odir}"

            outfilename = f"{odir}/{dname}.{args.format}"

            if not Path(odir).exists():
                log.debug("creating dir %s", odir)
                Path(odir).mkdir(parents=True)

            log.debug("rendering: %s -> %s...", pname, outfilename)

            if args.format == "pdf":
                save_pdf(outfilename, diagram)
            elif args.format == "svg":
                save_svg(outfilename, diagram)
            elif args.format == "png":
                save_png(outfilename, diagram)
            else:
                raise RuntimeError(f"Unknown file format: {args.format}")



from gaphas.handlemove import HandleMove
from gaphas.move import Move

from gaphor.diagram.tools.handlemove import (
    StickyAttachedHandleMove,
    sticky_attached_move,
)
from gaphor.SysML.blocks.proxyport import ProxyPortItem

HandleMove.register(ProxyPortItem, StickyAttachedHandleMove)
Move.register(ProxyPortItem, sticky_attached_move)



"""Adapters for the Property Editor.

To register property pages implemented in this module, it is imported in
gaphor.adapter package.
"""

from __future__ import annotations

import abc
from typing import Callable, List, Tuple, Type

import gaphas.item
from gaphas.segment import Segment
from gi.repository import GObject, Gtk

from gaphor.core import transactional
from gaphor.core.modeling import Diagram, Element, Presentation
from gaphor.i18n import translated_ui_string


class LabelValue(GObject.Object):
    __gtype_name__ = "LabelValue"

    def __init__(self, label, value):
        super().__init__()
        self._label = label
        self._value = value

    @GObject.Property(type=str)
    def label(self):
        return self._label

    @property
    def value(self):
        return self._value


def new_resource_builder(package, property_pages="propertypages"):
    def new_builder(*object_ids, signals=None):
        builder = Gtk.Builder(signals)
        builder.add_objects_from_string(
            translated_ui_string(package, f"{property_pages}.ui"), object_ids
        )
        return builder

    return new_builder


new_builder = new_resource_builder("gaphor.diagram")


class _PropertyPages:
    """Generic handler for property pages.

    Property pages are collected on type.
    """

    def __init__(self) -> None:
        self.pages: List[
            Tuple[Type[Element], Callable[[Element], PropertyPageBase]]
        ] = []

    def register(self, subject_type, func=None):
        def reg(func):
            self.pages.append((subject_type, func))
            return func

        return reg(func) if func else reg

    def __call__(self, subject):
        for subject_type, func in self.pages:
            if isinstance(subject, subject_type):
                yield func(subject)


PropertyPages = _PropertyPages()


class PropertyPageBase(abc.ABC):
    """A property page which can display itself in a notebook."""

    order = 100  # Order number, used for ordered display

    def __init__(self):
        super().__init__()

    @abc.abstractmethod
    def construct(self) -> Gtk.Widget | None:
        """Create the page (Gtk.Widget) that belongs to the Property page.

        Returns the page's toplevel widget (Gtk.Widget).
        """


def help_link(builder, help_widget, popover):
    """Show the help popover for a `Help` link in the property page."""

    def on_activate(*_args):
        builder.get_object(popover).set_visible(True)

    help = builder.get_object(help_widget)
    help.set_accessible_role(Gtk.AccessibleRole.BUTTON)
    click_handler = Gtk.GestureClick.new()
    click_handler.connect("released", on_activate)
    help.add_controller(click_handler)


def handler_blocking(widget, event_name, widget_handler):
    changed_id = widget.connect(event_name, widget_handler)

    def handler_wrapper(func):
        def _handler(event):
            widget.handler_block(changed_id)
            func(event)
            widget.handler_unblock(changed_id)

        return _handler

    return handler_wrapper


def _do_unparent(widget, _pspec, watcher):
    if not widget.props.parent:
        watcher.unsubscribe_all()


def unsubscribe_all_on_destroy(widget, watcher):
    widget.connect("notify::parent", _do_unparent, watcher)
    return widget


@PropertyPages.register(Diagram)
class NamePropertyPage(PropertyPageBase):
    """An adapter which works for any named item view.

    It also sets up a table view which can be extended.
    """

    order = 10

    def __init__(self, subject):
        assert subject is None or hasattr(subject, "name")
        super().__init__()
        self.subject = subject
        self.watcher = subject.watcher() if subject else None

    def construct(self):
        if not self.subject:
            return

        assert self.watcher
        builder = new_builder(
            "name-editor",
        )

        subject = self.subject

        entry = builder.get_object("name-entry")
        entry.set_text(subject and subject.name or "")

        type = builder.get_object("type-label")
        type.set_text(subject.__class__.__name__)

        @handler_blocking(entry, "changed", self._on_name_changed)
        def handler(event):
            if event.element is subject and event.new_value != entry.get_text():
                entry.set_text(event.new_value or "")

        self.watcher.watch("name", handler)

        return unsubscribe_all_on_destroy(
            builder.get_object("name-editor"), self.watcher
        )

    @transactional
    def _on_name_changed(self, entry):
        if self.subject.name != entry.get_text():
            self.subject.name = entry.get_text()


@PropertyPages.register(gaphas.item.Line)
class LineStylePage(PropertyPageBase):
    """Basic line style properties: color, orthogonal, etc."""

    order = 400

    def __init__(self, item):
        super().__init__()
        self.item = item
        self.horizontal_button: Gtk.Button

    def construct(self):
        builder = new_builder(
            "line-editor",
            signals={
                "rectilinear-changed": (self._on_orthogonal_change,),
                "orientation-changed": (self._on_horizontal_change,),
            },
        )

        rectilinear_button = builder.get_object("line-rectilinear")
        horizontal_button = builder.get_object("flip-orientation")

        self.horizontal_button = horizontal_button

        rectilinear_button.set_active(self.item.orthogonal)
        horizontal_button.set_active(self.item.horizontal)
        horizontal_button.set_sensitive(self.item.orthogonal)

        return builder.get_object("line-editor")

    @transactional
    def _on_orthogonal_change(self, button, gparam):
        if len(self.item.handles()) < 3:
            line_segment = Segment(self.item, self.item.diagram)
            line_segment.split_segment(0)
        active = button.get_active()
        self.item.orthogonal = active
        self.item.diagram.update_now((self.item,))
        self.horizontal_button.set_sensitive(active)

    @transactional
    def _on_horizontal_change(self, button, gparam):
        self.item.horizontal = button.get_active()
        self.item.diagram.update_now((self.item,))


@PropertyPages.register(Element)
class NotePropertyPage(PropertyPageBase):
    """A facility to add a little note/remark."""

    order = 300

    def __init__(self, subject):
        self.subject = subject.subject if isinstance(subject, Presentation) else subject
        self.watcher = self.subject and self.subject.watcher()

    def construct(self):
        subject = self.subject

        if not subject:
            return

        builder = new_builder("note-editor")
        text_view = builder.get_object("note")

        buffer = Gtk.TextBuffer()
        if subject.note:
            buffer.set_text(subject.note)
        text_view.set_buffer(buffer)

        @handler_blocking(buffer, "changed", self._on_body_change)
        def handler(event):
            if not text_view.props.has_focus:
                buffer.set_text(event.new_value or "")

        self.watcher.watch("note", handler)
        text_view.connect("destroy", self.watcher.unsubscribe_all)

        return builder.get_object("note-editor")

    @transactional
    def _on_body_change(self, buffer):
        self.subject.note = buffer.get_text(
            buffer.get_start_iter(), buffer.get_end_iter(), False
        )



"""This plugin extends Gaphor with XMI export functionality."""

import logging
from pathlib import Path

from gaphor.abc import ActionProvider, Service
from gaphor.core import action, gettext
from gaphor.plugins.xmiexport import exportmodel
from gaphor.ui.filedialog import save_file_dialog

logger = logging.getLogger(__name__)


class XMIExport(Service, ActionProvider):
    def __init__(self, element_factory, file_manager, export_menu):
        self.element_factory = element_factory
        self.file_manager = file_manager
        export_menu.add_actions(self)

    def shutdown(self):
        pass

    @action(
        name="file-export-xmi",
        label=gettext("Export as XMI"),
        tooltip=gettext("Export model as XMI (XML Model Interchange) format"),
    )
    def execute(self):
        def handler(filename):
            logger.debug(f"Exporting XMI model to: {filename}")
            export = exportmodel.XMIExport(self.element_factory)
            try:
                export.export(filename)
            except Exception as e:
                logger.error(f"Error while saving model to file {filename}: {e}")

        filename = (
            self.file_manager.filename.with_suffix(".xmi")
            if self.file_manager.filename
            else Path("model.xmi")
        )
        save_file_dialog(
            gettext("Export model as XMI file"),
            handler,
            filename=filename,
            extension=".xmi",
            filters=[(gettext("All XMI Files"), ".xmi", "text/xml")],
        )



"""Support functions for information flow

And by extension SysML item flow.
"""

from gaphor.core.format import format
from gaphor.diagram.presentation import Presentation
from gaphor.diagram.shapes import Text, cairo_state
from gaphor.UML.classes.association import get_center_pos
from gaphor.UML.recipes import stereotypes_str


def shape_information_flow(presentation: Presentation, attribute: str) -> list[Text]:
    return [
        Text(
            text=lambda: ", ".join(
                getattr(presentation.subject, attribute)[:].conveyed[:].name
            )
        ),
        # Also support SysML ItemFlow:
        Text(
            text=lambda: stereotypes_str(
                iflow[0].itemProperty.type,
                raaml_stereotype_workaround(iflow[0].itemProperty.type),
            )
            if (iflow := getattr(presentation.subject, attribute))
            else ""
        ),
        Text(
            text=lambda: format(iflow[0].itemProperty, type=True)
            if (iflow := getattr(presentation.subject, attribute))
            else ""
        ),
    ]


def watch_information_flow(
    presentation: Presentation, cast: str, attribute: str
) -> None:
    presentation.watch(f"subject[{cast}].{attribute}.informationSource")
    presentation.watch(f"subject[{cast}].{attribute}.conveyed.name")
    presentation.watch(f"subject[{cast}].{attribute}[ItemFlow].itemProperty.name")
    presentation.watch(f"subject[{cast}].{attribute}[ItemFlow].itemProperty.type.name")
    presentation.watch(
        f"subject[{cast}].{attribute}[ItemFlow].itemProperty.type.appliedStereotype.classifier.name"
    )


def draw_information_flow(presentation: Presentation, context, invert: bool) -> None:
    handles = presentation.handles()
    pos, angle = get_center_pos(handles)
    f = 1 if invert else -1
    with cairo_state(context.cairo) as cr:
        cr.translate(*pos)
        cr.rotate(angle)
        cr.move_to(0, 0)
        cr.line_to(12 * f, 8)
        cr.line_to(12 * f, -8)
        cr.fill()


def raaml_stereotype_workaround(element):
    """This is a temporary fix to ensure ItemFlow is showing proper stereotypes
    for Controller, Feedback and ControlAction from RAAML."""
    if not element:
        return ()

    name: str = type(element).__name__
    if name in {"Controller", "Feedback", "ControlAction"}:
        return (name[0].lower() + name[1:],)
    return ()



from gaphor.diagram.copypaste import copy, copy_presentation
from gaphor.UML import Activity, ActivityParameterNode
from gaphor.UML.actions.activity import ActivityParameterNodeItem
from gaphor.UML.actions.partition import PartitionItem
from gaphor.UML.copypaste import copy_named_element


@copy.register
def copy_partition_item(element: PartitionItem):
    yield element.id, copy_presentation(element)
    for partition in element.partition:
        yield from copy(partition)


@copy.register
def copy_activity(element: Activity):
    yield element.id, copy_named_element(element)
    for node in element.node:
        if isinstance(node, ActivityParameterNode):
            yield from copy(node)


@copy.register
def copy_activity_parameter_node(element: ActivityParameterNode):
    yield element.id, copy_named_element(element)
    yield from copy(element.parameter)


@copy.register
def copy_activity_parameter_node_item(element: ActivityParameterNodeItem):
    return []



from __future__ import annotations

import math
from collections.abc import Iterable, Iterator, Sequence
from dataclasses import replace
from enum import Enum
from math import pi
from typing import Callable, Protocol

from gaphas.geometry import Rectangle

from gaphor.core.modeling import DrawContext, Element, UpdateContext
from gaphor.core.modeling.diagram import lookup_attribute
from gaphor.core.styling import (
    JustifyContent,
    Number,
    Padding,
    Style,
    StyleNode,
    TextAlign,
    VerticalAlign,
    WhiteSpace,
    merge_styles,
)
from gaphor.core.styling.inherit import compute_inherited_style
from gaphor.core.styling.pseudo import compute_pseudo_element_style
from gaphor.diagram.text import Layout


class cairo_state:
    def __init__(self, cr):
        self._cr = cr

    def __enter__(self):
        self._cr.save()
        return self._cr

    def __exit__(self, _type, _value, _traceback):
        self._cr.restore()


def stroke(context: DrawContext, fill: bool, dash=True):
    style = context.style
    cr = context.cairo
    fill_color = style.get("background-color")
    if fill and fill_color:
        with cairo_state(cr):
            cr.set_source_rgba(*fill_color)
            cr.fill_preserve()

    with cairo_state(cr):
        if stroke := style.get("color"):
            cr.set_source_rgba(*stroke)
        if line_width := style.get("line-width", 2):
            cr.set_line_width(line_width)
        if dash:
            cr.set_dash(style.get("dash-style", ()), 0)
        cr.stroke()


def draw_border(box, context: DrawContext, bounding_box: Rectangle):
    cr = context.cairo
    x, y, width, height = bounding_box
    d = min(context.style.get("border-radius", 0), width / 2.0, height / 2.0)

    cr.move_to(x, y + d)
    cr.set_dash(context.style.get("dash-style", ()), 0)
    if d:
        x1 = x + width
        y1 = y + height
        cr.arc(x + d, y + d, d, pi, 1.5 * pi)
        cr.line_to(x1 - d, y)
        cr.arc(x1 - d, y + d, d, 1.5 * pi, 0)
        cr.line_to(x1, y1 - d)
        cr.arc(x1 - d, y1 - d, d, 0, 0.5 * pi)
        cr.line_to(x + d, y1)
        cr.arc(x + d, y1 - d, d, 0.5 * pi, pi)
    else:
        cr.rectangle(x, y, width, height)

    cr.close_path()

    stroke(context, fill=True)


def draw_top_separator(box: Box, context: DrawContext, bounding_box: Rectangle):
    x, y, w, _h = bounding_box
    cr = context.cairo
    cr.move_to(x, y)
    cr.line_to(x + w, y)

    stroke(context, fill=False)


def draw_left_separator(box: Box, context: DrawContext, bounding_box: Rectangle):
    x, y, _w, h = bounding_box
    cr = context.cairo
    cr.move_to(x, y)
    cr.line_to(x, y + h)

    stroke(context, fill=False)


def draw_ellipse(box: Box, context: DrawContext, bounding_box: Rectangle):
    ellipse(context.cairo, *bounding_box)

    stroke(context, fill=True)


def ellipse(cr, x, y, w, h, dc=None):
    if dc is None:
        dc = (4 / 3) * math.tan(math.pi / 8)
    rx = x + w / 2.0
    ry = y + h / 2.0
    x1 = x + w
    y1 = y + h
    tw = w * dc / 2
    th = h * dc / 2

    # curve: control_a, control_b, end
    cr.move_to(x, ry)
    cr.curve_to(x, ry - th, rx - tw, y, rx, y)
    cr.curve_to(rx + tw, y, x1, ry - th, x1, ry)
    cr.curve_to(x1, ry + th, rx + tw, y1, rx, y1)
    cr.curve_to(rx - tw, y1, x, ry + th, x, ry)

    cr.close_path()


def draw_default_head(context: DrawContext):
    """Default head drawer: move cursor to the first handle."""
    context.cairo.move_to(0, 0)


def draw_default_tail(context: DrawContext):
    """Default tail drawer: draw line to the last handle."""
    context.cairo.line_to(0, 0)


def draw_arrow_head(context: DrawContext):
    cr = context.cairo
    cr.save()
    cr.set_dash((), 0)
    cr.move_to(15, -6)
    cr.line_to(0, 0)
    cr.line_to(15, 6)
    stroke(context, fill=False, dash=False)
    cr.restore()
    cr.move_to(0, 0)


def draw_arrow_tail(context: DrawContext):
    cr = context.cairo
    cr.line_to(0, 0)
    stroke(context, fill=False)
    cr.save()
    cr.set_dash((), 0)
    cr.move_to(15, -6)
    cr.line_to(0, 0)
    cr.line_to(15, 6)
    stroke(context, fill=False, dash=False)
    cr.restore()


def draw_diamond(
    context: DrawContext, x1: float, x2: float, y1: float, y2: float
) -> None:
    """Draw a diamond."""
    cr = context.cairo
    center_x = x1 + (x2 - x1) / 2.0
    center_y = y1 + (y2 - y1) / 2.0
    cr.move_to(x1, center_y)
    cr.line_to(center_x, y2)
    cr.line_to(x2, center_y)
    cr.line_to(center_x, y1)
    cr.line_to(x1, center_y)
    stroke(context, fill=True)


class Orientation(Enum):
    VERTICAL = "v"
    HORIZONTAL = "h"


def rectangle_shrink(rect: Rectangle | None, padding: Padding) -> Rectangle:
    if rect is None:
        return Rectangle()

    top, right, bottom, left = padding
    return Rectangle(
        rect.x + left,
        rect.y + top,
        rect.width - left - right,
        rect.height - top - bottom,
    )


class Shape(Iterable, Protocol):
    def size(
        self, context: UpdateContext, bounding_box: Rectangle | None = None
    ) -> tuple[Number, Number]:
        ...

    def draw(self, context: DrawContext, bounding_box: Rectangle) -> None:
        ...


DEFAULT_PADDING = (0, 0, 0, 0)


class Box:
    """A box like shape.

    Style properties:
    - min-height
    - min-width
    - padding: a tuple (top, right, bottom, left)
    - justify-content: alignment of child shapes
    """

    def __init__(
        self,
        *children: Shape,
        orientation: Orientation = Orientation.VERTICAL,
        style: Style | None = None,
        draw: Callable[[Box, DrawContext, Rectangle], None] | None = None,
    ):
        if style is None:
            style = {}
        self.children = children
        self.sizes: list[tuple[Number, Number]] = []
        self._orientation = orientation
        self._inline_style = style
        self._draw_border = draw

    def __iter__(self):
        return iter(self.children)

    def __getitem__(self, index):
        return self.children[index]

    def size(self, context: UpdateContext, bounding_box: Rectangle | None = None):
        style = merge_styles(context.style, self._inline_style)
        min_width = style.get("min-width", 0)
        min_height = style.get("min-height", 0)
        padding = style.get("padding", DEFAULT_PADDING)
        new_bounds = rectangle_shrink(bounding_box, padding)
        self.sizes = sizes = [c.size(context, new_bounds) for c in self.children]
        if sizes:
            widths, heights = list(zip(*sizes))
            is_vertical = self._orientation == Orientation.VERTICAL
            padding_top, padding_right, padding_bottom, padding_left = padding
            return (
                max(
                    min_width,
                    (max(widths) if is_vertical else sum(widths))
                    + padding_right
                    + padding_left,
                ),
                max(
                    min_height,
                    (sum(heights) if is_vertical else max(heights))
                    + padding_top
                    + padding_bottom,
                ),
            )
        return min_width, min_height

    def draw(self, context: DrawContext, bounding_box: Rectangle):
        if self._orientation == Orientation.VERTICAL:
            self.draw_vertical(context, bounding_box)
        else:
            self.draw_horizontal(context, bounding_box)

    def draw_vertical(self, context: DrawContext, bounding_box: Rectangle):
        style = merge_styles(context.style, self._inline_style)
        new_context = replace(context, style=style)
        padding_top, padding_right, padding_bottom, padding_left = style.get(
            "padding", DEFAULT_PADDING
        )
        sizes = self.sizes

        justify_content = (
            style.get("justify-content")
            or {
                VerticalAlign.TOP: JustifyContent.START,
                VerticalAlign.MIDDLE: JustifyContent.CENTER,
                VerticalAlign.BOTTOM: JustifyContent.END,
            }[style.get("vertical-align", VerticalAlign.MIDDLE)]
        )

        if justify_content is JustifyContent.STRETCH and sizes:
            height = bounding_box.height
            avg_height = height / len(sizes)
            oversized = [h for _w, h in sizes if h > avg_height]
            d = len(sizes) - len(oversized)
            avg_height = (height - sum(oversized)) / d if d > 0 else 0
        else:
            height = sum(h for _w, h in sizes)
            avg_height = 0

        if justify_content is JustifyContent.CENTER:
            y = (
                bounding_box.y
                + padding_top
                + (max(height, bounding_box.height - padding_top) - height) / 2
            )
        elif justify_content is JustifyContent.END:
            y = bounding_box.y + bounding_box.height - height - padding_bottom
        else:
            y = bounding_box.y + padding_top

        if self._draw_border:
            self._draw_border(self, new_context, bounding_box)

        x = bounding_box.x + padding_left
        w = bounding_box.width - padding_right - padding_left
        if self.children:
            last_child = self.children[-1]
            for c, (_w, h) in zip(self.children, sizes):
                if c is last_child and justify_content is JustifyContent.START:
                    h = bounding_box.height - y
                elif h < avg_height:
                    h = avg_height
                c.draw(context, Rectangle(x, y, w, h))
                y += h

    def draw_horizontal(self, context: DrawContext, bounding_box: Rectangle):
        style = merge_styles(context.style, self._inline_style)
        new_context = replace(context, style=style)
        padding_top, padding_right, padding_bottom, padding_left = style.get(
            "padding", DEFAULT_PADDING
        )
        sizes = self.sizes

        justify_content = style.get("justify-content", JustifyContent.CENTER)

        if justify_content is JustifyContent.STRETCH and sizes:
            width = bounding_box.width
            avg_width = width / len(sizes)
            oversized = [w for w, _h in sizes if w > avg_width]
            d = len(sizes) - len(oversized)
            avg_width = (width - sum(oversized)) / d if d > 0 else 0
        else:
            width = sum(w for w, _h in sizes)
            avg_width = 0

        if justify_content is JustifyContent.CENTER:
            x = (
                bounding_box.x
                + padding_left
                + (max(width, bounding_box.width - padding_left) - width) / 2
            )
        elif justify_content is JustifyContent.END:
            x = bounding_box.x + bounding_box.width - width - padding_right
        else:
            x = bounding_box.x + padding_left

        if self._draw_border:
            self._draw_border(self, new_context, bounding_box)

        y = bounding_box.y + padding_top
        h = bounding_box.height - padding_bottom - padding_top
        if self.children:
            last_child = self.children[-1]
            for c, (w, _h) in zip(self.children, sizes):
                if c is last_child and justify_content is JustifyContent.START:
                    w = bounding_box.height - x
                elif w < avg_width:
                    w = avg_width
                c.draw(context, Rectangle(x, y, w, h))
                x += w


class BoundedBox(Box):
    """A Box.

    It keeps track of the latest bounding box used for drawing.
    """

    def __init__(
        self,
        *children: Shape,
        style: Style | None = None,
        draw: Callable[[Box, DrawContext, Rectangle], None] | None = None,
    ):
        if style is None:
            style = {}
        super().__init__(*children, style=style, draw=draw)
        self.bounding_box = Rectangle()

    def __iter__(self):
        return iter(self.children)

    def draw(self, context: DrawContext, bounding_box: Rectangle):
        self.bounding_box = bounding_box
        return super().draw(context, bounding_box)

    def __contains__(self, pos_or_rect):
        return pos_or_rect in self.bounding_box


class IconBox:
    """A special type of box: the icon element is given the full width/height
    and all other shapes are drawn below the main icon shape.

    Style properties:
    - min-height
    - min-width
    - vertical-spacing: spacing between icon and children
    - padding: a tuple (top, right, bottom, left)
    """

    def __init__(self, icon: Shape, *children: Shape, style: Style | None = None):
        if style is None:
            style = {}
        self.icon = icon
        self.children = children
        self.sizes: list[tuple[Number, Number]] = []
        self._inline_style = style

    def __iter__(self):
        return iter((self.icon, *self.children))

    def size(self, context: UpdateContext, bounding_box: Rectangle | None = None):
        style = merge_styles(context.style, self._inline_style)
        min_width = style.get("min-width", 0)
        min_height = style.get("min-height", 0)
        padding = style.get("padding", DEFAULT_PADDING)

        new_bounds = rectangle_shrink(bounding_box, padding)
        width, height = self.icon.size(context, new_bounds)

        new_bounds.expand(new_bounds.width * 1.32)
        self.sizes = [c.size(context, new_bounds) for c in self.children]

        padding_top, padding_right, padding_bottom, padding_left = padding
        return (
            max(min_width, width + padding_right + padding_left),
            max(min_height, height + padding_top + padding_bottom),
        )

    def child_pos(self, style: Style, bounding_box: Rectangle) -> Rectangle:
        if not self.sizes:
            return Rectangle()

        text_align = style.get("text-align", TextAlign.CENTER)
        vertical_align = style.get("vertical-align", VerticalAlign.BOTTOM)
        vertical_spacing = style.get("vertical-spacing", 0)  # should be margin?

        ws, hs = list(zip(*self.sizes))
        max_w = max(ws)
        total_h = sum(hs)

        if text_align == TextAlign.CENTER:
            x = bounding_box.x + (bounding_box.width - max_w) / 2
        elif text_align == TextAlign.LEFT:
            x = bounding_box.x - max_w - vertical_spacing
        elif text_align == TextAlign.RIGHT:
            x = bounding_box.x + bounding_box.width + vertical_spacing

        if vertical_align == VerticalAlign.BOTTOM:
            y = bounding_box.y + bounding_box.height + vertical_spacing
        elif vertical_align == VerticalAlign.MIDDLE:
            y = bounding_box.y + (bounding_box.height - total_h) / 2
        elif vertical_align == VerticalAlign.TOP:
            y = bounding_box.y - total_h - vertical_spacing
        return Rectangle(
            x,
            y,
            max_w,
            total_h,
        )

    def draw(self, context: DrawContext, bounding_box: Rectangle):
        style = merge_styles(context.style, self._inline_style)
        new_context = replace(context, style=style)
        self.icon.draw(
            new_context,
            rectangle_shrink(bounding_box, style.get("padding", DEFAULT_PADDING)),
        )

        cx, cy, max_w, total_h = self.child_pos(style, bounding_box)
        for c, (cw, ch) in zip(self.children, self.sizes):
            c.draw(context, Rectangle(cx + (max_w - cw) / 2, cy, cw, ch))
            cy += ch


class Text:
    def __init__(self, text: str | Callable[[], str], style: Style | None = None):
        if style is None:
            style = {}
        self._text = text if callable(text) else lambda: text
        self._inline_style = style
        self._layout = Layout()

    def __iter__(self):
        return iter(())

    def text(self, style: Style | None = None):
        try:
            t = self._text()
        except AttributeError:
            t = ""

        if (
            style
            and (after := compute_pseudo_element_style(style, "after"))
            and (content := after.get("content"))
        ):
            return f"{t}{content}"
        return t

    def size(self, context: UpdateContext, bounding_box: Rectangle | None = None):
        style = merge_styles(context.style, self._inline_style)
        min_w = style.get("min-width", 0)
        min_h = style.get("min-height", 0)
        text_align = style.get("text-align", TextAlign.CENTER)
        white_space = style.get("white-space", WhiteSpace.NORMAL)

        layout = self._layout
        layout.set(
            text=self.text(style),
            font=style,
            width=bounding_box.width
            if bounding_box and white_space == WhiteSpace.NORMAL
            else -1,
            text_align=text_align,
        )
        width, height = layout.size()
        padding_top, padding_right, padding_bottom, padding_left = style.get(
            "padding", DEFAULT_PADDING
        )
        return (
            max(min_w, width + padding_right + padding_left),
            max(min_h, height + padding_top + padding_bottom),
        )

    def draw(self, context: DrawContext, bounding_box: Rectangle):
        """Draw the text, return the location and size."""
        style = merge_styles(context.style, self._inline_style)
        min_w = max(style.get("min-width", 0), bounding_box.width)
        min_h = max(style.get("min-height", 0), bounding_box.height)
        text_box = rectangle_shrink(bounding_box, style.get("padding", DEFAULT_PADDING))

        with cairo_state(context.cairo) as cr:
            if text_color := style.get("text-color"):
                cr.set_source_rgba(*text_color)
            elif color := style.get("color"):
                cr.set_source_rgba(*color)

            layout = self._layout
            cr.move_to(text_box.x, text_box.y)
            layout.set(font=style)
            layout.show_layout(cr, text_box.width, default_size=(min_w, min_h))


class CssNode:
    """Create custom styling based on the active style and
    an element/class declaration.
    """

    def __init__(
        self,
        name: str,
        element: Element | None,
        child: Shape,
        style: Style | None = None,
    ):
        self._name = name
        self._element = element
        self._child = child
        self._inline_style = style if style else {}

    def __iter__(self):
        return iter((self._child,))

    def style_node(self, parent: Style | StyleNode) -> StyleNode:
        return StyledCssNode(parent, self)

    def size(
        self, context: UpdateContext, bounding_box: Rectangle | None = None
    ) -> tuple[Number, Number]:
        style = compute_inherited_style(context.style, self.style_node(context.style))
        new_context = replace(context, style=style)

        return self._child.size(new_context, bounding_box)

    def draw(self, context: DrawContext, bounding_box: Rectangle):
        style = compute_inherited_style(context.style, self.style_node(context.style))
        new_context = replace(context, style=style)

        self._child.draw(new_context, bounding_box)


class StyledCssNode:
    def __init__(self, parent: Style | StyleNode, shape: CssNode):
        self._parent: StyleNode | None = (
            parent.get("-gaphor-style-node") if isinstance(parent, dict) else parent  # type: ignore[assignment]
        )
        self._shape = shape
        self.pseudo: str | None = None
        self.dark_mode = self._parent.dark_mode if self._parent else None

    def name(self) -> str:
        return self._shape._name

    def parent(self) -> StyleNode | None:
        return self._parent

    def children(self) -> Iterator[StyleNode]:
        return (node.style_node(self) for node in traverse_css_nodes(self._shape))

    def attribute(self, name: str) -> str | None:
        if not self._shape._element:
            return None
        return lookup_attribute(self._shape._element, name)

    def state(self) -> Sequence[str]:
        return ()

    def __hash__(self):
        return hash((self._parent, self._shape))

    def __eq__(self, other):
        return (
            isinstance(other, StyledCssNode)
            and self._parent == other._parent
            and self._shape == other._shape
        )


def traverse_css_nodes(shape: Shape) -> Iterator[CssNode]:
    for s in shape:
        if isinstance(s, CssNode):
            yield s
        else:
            yield from traverse_css_nodes(s)



from functools import singledispatch

from gaphor.core.modeling import Element


@singledispatch
def deletable(element: Element) -> bool:
    """Determine if a single element can safely be deleted."""
    return True



from __future__ import annotations

from gi.repository import Gio, GObject, Gtk

from gaphor import UML
from gaphor.core import event_handler, gettext, transactional
from gaphor.core.format import format, parse
from gaphor.core.modeling import AssociationUpdated
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    handler_blocking,
    help_link,
    new_resource_builder,
    unsubscribe_all_on_destroy,
)
from gaphor.diagram.propertypages import (
    new_builder as diagram_new_builder,
)
from gaphor.UML.actions.activity import ActivityItem, ActivityParameterNodeItem
from gaphor.UML.propertypages import (
    TypedElementPropertyPage,
    create_list_store,
    list_item_factory,
    list_view_activated,
    list_view_key_handler,
    text_field_handlers,
    update_list_store,
)

new_builder = new_resource_builder("gaphor.UML.actions")


class ActivityParameterNodeView(GObject.Object):
    def __init__(self, node: UML.ActivityParameterNode | None, activity: UML.Activity):
        super().__init__()
        self.node = node
        self.activity = activity

    editing = GObject.Property(type=bool, default=False)

    @GObject.Property(type=str)
    def parameter(self) -> str:
        return format(self.node.parameter) if self.node else ""

    @parameter.setter  # type: ignore[no-redef]
    @transactional
    def parameter(self, value):
        if not self.node:
            if not value:
                return

            model = self.activity.model
            node = model.create(UML.ActivityParameterNode)
            node.parameter = model.create(UML.Parameter)
            self.node = node
            self.activity.node = node
        parse(self.node.parameter, value)

    def start_editing(self):
        self.editing = True

    def empty(self):
        return not self.node

    def unlink(self):
        if self.node:
            self.node.unlink()

    def swap(self, item1, item2):
        return self.activity.node.swap(item1.node, item2.node)


def activity_parameter_node_model(activity: UML.Activity) -> Gio.ListModel:
    return create_list_store(
        ActivityParameterNodeView,
        (
            node
            for node in activity.node
            if isinstance(node, UML.ActivityParameterNode) and node.parameter
        ),
        lambda node: ActivityParameterNodeView(node, activity),
    )


def update_activity_parameter_node_model(
    store: Gio.ListStore, activity: UML.Activity
) -> Gio.ListStore:
    return update_list_store(
        store,
        lambda item: item.node,
        (
            node
            for node in activity.node
            if isinstance(node, UML.ActivityParameterNode) and node.parameter
        ),
        lambda node: ActivityParameterNodeView(node, activity),
    )


@PropertyPages.register(ActivityItem)
class ActivityItemPage(PropertyPageBase):
    order = 40

    def __init__(self, item: ActivityItem):
        self.item = item
        self.watcher = item.subject and item.subject.watcher()

    def construct(self):
        subject = self.item.subject

        if not subject:
            return

        builder = new_builder(
            "activity-editor",
            "parameters-info",
            signals={
                "parameter-activated": (list_view_activated,),
                "parameter-key-pressed": (list_view_key_handler,),
                "parameters-info-clicked": (self.on_parameters_info_clicked,),
            },
        )

        self.info = builder.get_object("parameters-info")
        help_link(builder, "parameters-info-icon", "parameters-info")

        column_view: Gtk.ListView = builder.get_object("parameter-list")

        for column, factory in zip(
            column_view.get_columns(),
            [
                list_item_factory(
                    "text-field-cell.ui",
                    klass=ActivityParameterNodeView,
                    attribute=ActivityParameterNodeView.parameter,
                    placeholder_text=gettext("New Parameter…"),
                    signal_handlers=text_field_handlers("parameter"),
                ),
            ],
        ):
            column.set_factory(factory)

        self.model = activity_parameter_node_model(subject)
        selection = Gtk.SingleSelection.new(self.model)
        column_view.set_model(selection)

        if self.watcher:
            self.watcher.watch("node", self.on_nodes_changed)

        return unsubscribe_all_on_destroy(
            builder.get_object("activity-editor"), self.watcher
        )

    @event_handler(AssociationUpdated)
    def on_nodes_changed(self, event):
        update_activity_parameter_node_model(self.model, self.item.subject)

    def on_parameters_info_clicked(self, image, event):
        self.info.set_visible(True)


@PropertyPages.register(UML.ActivityParameterNode)
class ActivityParameterNodeNamePropertyPage(PropertyPageBase):
    """An adapter which works for any named item view.

    It also sets up a table view which can be extended.
    """

    order = 10

    def __init__(self, subject):
        assert subject is None or hasattr(subject, "name")
        super().__init__()
        self.subject = subject
        self.watcher = subject.watcher() if subject else None

    def construct(self):
        if not self.subject:
            return

        assert self.watcher
        builder = diagram_new_builder(
            "name-editor",
        )

        subject = self.subject

        entry = builder.get_object("name-entry")
        entry.set_text(subject and subject.parameter and subject.parameter.name or "")

        @handler_blocking(entry, "changed", self._on_name_changed)
        def handler(event):
            if event.element is subject and event.new_value != entry.get_text():
                entry.set_text(event.new_value or "")

        self.watcher.watch("parameter.name", handler)

        return unsubscribe_all_on_destroy(
            builder.get_object("name-editor"), self.watcher
        )

    @transactional
    def _on_name_changed(self, entry):
        if self.subject.parameter.name != entry.get_text():
            self.subject.parameter.name = entry.get_text()


@PropertyPages.register(ActivityParameterNodeItem)
class ActivityParameterNodeTypePropertyPage(TypedElementPropertyPage):
    @property
    def typed_element(self):
        return self.item.subject.parameter


@PropertyPages.register(ActivityParameterNodeItem)
class ActivityParameterNodeDirectionPropertyPage(PropertyPageBase):
    DIRECTION = UML.Parameter.direction.values
    order = 40

    def __init__(self, item):
        super().__init__()
        self.item = item

    def construct(self):
        if not (self.item.subject and self.item.subject.parameter):
            return

        builder = new_builder(
            "parameter-direction-editor",
            signals={
                "parameter-direction-changed": (self._on_parameter_direction_changed,),
                "show-direction-changed": (self._on_show_direction_changed,),
            },
        )

        direction = builder.get_object("parameter-direction")
        direction.set_selected(
            self.DIRECTION.index(self.item.subject.parameter.direction)
        )

        show_direction = builder.get_object("show-direction")
        show_direction.set_active(self.item.show_direction)

        return builder.get_object("parameter-direction-editor")

    @transactional
    def _on_parameter_direction_changed(self, dropdown, _pspec):
        self.item.subject.parameter.direction = self.DIRECTION[dropdown.get_selected()]

    @transactional
    def _on_show_direction_changed(self, button, _gspec):
        self.item.show_direction = button.get_active()



"""The action definition for the UML toolbox."""

from gaphor.diagram.diagramtoolbox import (
    DiagramType,
    DiagramTypes,
    ElementCreateInfo,
    ToolboxDefinition,
    general_tools,
)
from gaphor.i18n import i18nize
from gaphor.UML.actions.actionstoolbox import actions
from gaphor.UML.classes.classestoolbox import classes
from gaphor.UML.deployments.deploymentstoolbox import deployments
from gaphor.UML.interactions.interactionstoolbox import interactions
from gaphor.UML.profiles.profilestoolbox import profiles
from gaphor.UML.states.statestoolbox import states
from gaphor.UML.uml import (
    Activity,
    Actor,
    Artifact,
    Class,
    Component,
    DataType,
    Device,
    Enumeration,
    InstanceSpecification,
    Interaction,
    Interface,
    Node,
    Package,
    Profile,
    StateMachine,
    Stereotype,
    UseCase,
)
from gaphor.UML.usecases.usecasetoolbox import use_cases

uml_toolbox_actions: ToolboxDefinition = (
    general_tools,
    classes,
    deployments,
    actions,
    interactions,
    states,
    use_cases,
    profiles,
)

uml_diagram_types: DiagramTypes = (
    DiagramType("cls", i18nize("New Class Diagram"), (classes,)),
    DiagramType("pkg", i18nize("New Package Diagram"), (classes,)),
    DiagramType("cmp", i18nize("New Component Diagram"), (classes,)),
    DiagramType("dep", i18nize("New Deployment Diagram"), (deployments,)),
    DiagramType("act", i18nize("New Activity Diagram"), (actions,)),
    DiagramType("sd", i18nize("New Sequence Diagram"), (interactions,)),
    DiagramType("com", i18nize("New Communication Diagram"), (interactions,)),
    DiagramType("stm", i18nize("New State Machine Diagram"), (states,)),
    DiagramType("uc", i18nize("New Use Case Diagram"), (use_cases,)),
    DiagramType("prf", i18nize("New Profile Diagram"), (profiles,)),
)

uml_element_types = (
    ElementCreateInfo("activity", i18nize("New Activity"), Activity, (Package,)),
    ElementCreateInfo("actor", i18nize("New Actor"), Actor, (Package,)),
    ElementCreateInfo("artifact", i18nize("New Artifact"), Artifact, (Package,)),
    ElementCreateInfo("class", i18nize("New Class"), Class, (Package,)),
    ElementCreateInfo("component", i18nize("New Component"), Component, (Package,)),
    ElementCreateInfo("datatype", i18nize("New Data Type"), DataType, (Package,)),
    ElementCreateInfo("device", i18nize("New Device"), Device, (Package,)),
    ElementCreateInfo(
        "enumeration", i18nize("New Enumeration"), Enumeration, (Package,)
    ),
    ElementCreateInfo(
        "instancespecification",
        i18nize("New Instance Specification"),
        InstanceSpecification,
        (Package,),
    ),
    ElementCreateInfo(
        "interaction", i18nize("New Interaction"), Interaction, (Package,)
    ),
    ElementCreateInfo("interface", i18nize("New Interface"), Interface, (Package,)),
    ElementCreateInfo("node", i18nize("New Node"), Node, (Package,)),
    ElementCreateInfo("profile", i18nize("New Profile"), Profile, (Package,)),
    ElementCreateInfo(
        "statemachine", i18nize("New State Machine"), StateMachine, (Package,)
    ),
    ElementCreateInfo("stereotype", i18nize("New Stereotype"), Stereotype, (Package,)),
    ElementCreateInfo("usecase", i18nize("New Use Case"), UseCase, (Package,)),
)



from __future__ import annotations

from gaphas.types import Pos

from gaphor import UML
from gaphor.core.modeling.element import Element
from gaphor.core.modeling.properties import attribute
from gaphor.core.styling import JustifyContent
from gaphor.diagram.presentation import Classified, ElementPresentation, text_name
from gaphor.diagram.shapes import Box, Text, draw_border
from gaphor.diagram.support import represents
from gaphor.UML.classes.stereotype import stereotype_compartments, stereotype_watches
from gaphor.UML.recipes import stereotypes_str
from gaphor.UML.states.region import region_compartment


@represents(UML.StateMachine)
class StateMachineItem(Classified, ElementPresentation[UML.StateMachine]):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)
        self._region_boxes = []
        self.watch("show_stereotypes", self.update_shapes)
        self.watch("show_regions", self.update_shapes)
        self.watch("subject[NamedElement].name")
        self.watch("subject[StateMachine].region.name")
        self.watch("subject[StateMachine].region", self.update_shapes)
        stereotype_watches(self)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)
    show_regions: attribute[int] = attribute("show_regions", int, default=True)

    def update_shapes(self, event=None):
        self._region_boxes = (
            list(region_compartment(self.subject)) if self.show_regions else []
        )
        self.shape = Box(
            Box(
                Text(
                    text=lambda: stereotypes_str(
                        self.subject, [self.diagram.gettext("statemachine")]
                    ),
                ),
                text_name(self),
                style={"padding": (4, 4, 4, 4)},
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            Box(
                *(self._region_boxes),
                style={"justify-content": JustifyContent.STRETCH},
            ),
            style={
                "justify-content": JustifyContent.START
                if (self.diagram and self.children) or self.show_regions
                else JustifyContent.CENTER,
            },
            draw=draw_border,
        )

    def subject_at_point(self, pos: Pos) -> Element | None:
        return next(
            (
                region
                for region, bounds in zip(self.subject.region, self._region_boxes)
                if pos in bounds
            ),
            self.subject,
        )



"""Copy / Paste functionality."""

from __future__ import annotations

from typing import Callable, Collection

from gi.repository import Gdk, GLib, GObject

from gaphor.abc import ActionProvider, Service
from gaphor.core import Transaction, action
from gaphor.core.modeling import Diagram, Presentation
from gaphor.diagram.copypaste import Opaque, copy_full, paste_full, paste_link


class CopyBuffer(GObject.Object):
    def __init__(self, buffer):
        super().__init__()
        self.buffer = buffer

    buffer = GObject.Property(type=object)


class CopyService(Service, ActionProvider):
    """Copy/Cut/Paste functionality required a lot of thinking:

    Store a list of DiagramItems that have to be copied in a global
    'copy-buffer'.

    - In order to make copy/paste work, the load/save functions should be
      generalized to allow a subset to be saved/loaded (which is needed
      anyway for exporting/importing stereotype Profiles).
    - How much data should be saved? An example use case is to copy a diagram
      item, remove it (the underlying UML element is removed), and then paste
      the copied item. The diagram should act as if we have placed a copy of
      the removed item on the diagram and make the UML element visible again.
    """

    def __init__(self, event_manager, element_factory, diagrams):
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.diagrams = diagrams

        self.clipboard = Gdk.Display.get_default().get_primary_clipboard()

    def shutdown(self):
        pass

    def copy(self, items: Collection[Presentation]) -> None:
        if items:
            copy_buffer = copy_full(items, self.element_factory.lookup)
            v = GObject.Value(CopyBuffer.__gtype__, CopyBuffer(buffer=copy_buffer))
            self.clipboard.set_content(Gdk.ContentProvider.new_for_value(v))

    def paste_link(
        self,
        diagram: Diagram,
        callback: Callable[[set[Presentation]], None] | None = None,
    ) -> None:
        self._paste(diagram, paste_link, callback)

    def paste_full(
        self,
        diagram: Diagram,
        callback: Callable[[set[Presentation]], None] | None = None,
    ) -> None:
        self._paste(diagram, paste_full, callback)

    def _paste(
        self,
        diagram: Diagram,
        paster: Callable[[Opaque, Diagram], set[Presentation]],
        callback: Callable[[set[Presentation]], None] | None,
    ) -> None:
        def on_paste(_source_object, result):
            try:
                copy_buffer = self.clipboard.read_value_finish(result)
            except GLib.GError as e:
                if str(e).startswith("g-io-error-quark:"):
                    return
                raise

            with Transaction(self.event_manager):
                # Create new id's that have to be used to create the items:
                new_items = paster(copy_buffer.buffer, diagram)

                # move pasted items a bit, so user can see result of his action :)
                for item in new_items:
                    if item.parent not in new_items:
                        item.matrix.translate(10, 10)

            if callback:
                callback(new_items)

        self.clipboard.read_value_async(
            CopyBuffer.__gtype__,
            io_priority=GLib.PRIORITY_DEFAULT,
            cancellable=None,
            callback=on_paste,
        )

    @action(
        name="edit-copy",
        shortcut="<Primary>c",
    )
    def copy_action(self):
        view = self.diagrams.get_current_view()
        if view.is_focus():
            if items := view.selection.selected_items:
                self.copy(items)
            else:
                return

    @action(name="edit-cut", shortcut="<Primary>x")
    def cut_action(self):
        view = self.diagrams.get_current_view()
        if view.is_focus():
            items = view.selection.selected_items
            if not items:
                return

            self.copy(items)

            with Transaction(self.event_manager):
                for i in list(items):
                    i.unlink()

    @action(name="edit-paste-link", shortcut="<Primary>v")
    def paste_link_action(self):
        self._paste_action(paste_link)

    @action(name="edit-paste-full", shortcut="<Primary><Shift>v")
    def paste_full_action(self):
        self._paste_action(paste_full)

    def _paste_action(self, paster):
        view = self.diagrams.get_current_view()
        diagram = self.diagrams.get_current_diagram()
        if not (view and view.is_focus()):
            return

        def select_items(new_items):
            selection = view.selection
            selection.unselect_all()
            selection.select_items(*new_items)

        self._paste(diagram, paster, select_items)



"""
An item representing a diagram.
"""

from gaphor.core.modeling.diagram import Diagram
from gaphor.diagram.presentation import ElementPresentation, Named
from gaphor.diagram.shapes import Box, IconBox, Text, draw_border, stroke
from gaphor.diagram.support import represents
from gaphor.diagram.text import FontWeight


@represents(Diagram)
class DiagramItem(ElementPresentation, Named):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=30, height=30)
        for h in self.handles():
            h.movable = False

        self.shape = IconBox(
            Box(
                Box(draw=draw_diagram),
                style={
                    "padding": (4, 4, 4, 4),
                    "border-radius": 4,
                },
                draw=draw_border,
            ),
            Text(
                text=lambda: self.subject and self.subject.diagramType or "",
                style={
                    "font-weight": FontWeight.BOLD,
                },
            ),
            Text(text=lambda: self.subject and self.subject.name or ""),
        )

        self.watch("subject[Diagram].name")
        self.watch("subject[Diagram].diagramType")


def draw_diagram(box, context, bounding_box):
    cr = context.cairo

    cr.rectangle(5, 13, 8, 12)
    cr.rectangle(18, 5, 8, 12)
    cr.move_to(17, 9)
    cr.line_to(8, 9)
    cr.line_to(8, 12)
    stroke(context, fill=True)



from __future__ import annotations

from unicodedata import normalize

from gi.repository import Gio, GObject, Pango

from gaphor import UML
from gaphor.core.format import format
from gaphor.core.modeling import Diagram, Element
from gaphor.diagram.iconname import icon_name
from gaphor.i18n import gettext

_no_value = object()


class TreeItem(GObject.Object):
    def __init__(self, element: Element | None):
        super().__init__()
        self.element = element
        if element:
            self.sync()

    icon = GObject.Property(type=str)
    icon_visible = GObject.Property(type=bool, default=False)

    readonly_text = GObject.Property(type=str)
    attributes = GObject.Property(type=Pango.AttrList)
    editing = GObject.Property(type=bool, default=False)
    can_edit = GObject.Property(type=bool, default=True)

    def read_only(self):
        return not self.element or not hasattr(self.element, "name")

    @GObject.Property(type=str)
    def editable_text(self):
        return "" if self.read_only() else (self.element.name or "")

    @editable_text.setter  # type: ignore[no-redef]
    def editable_text(self, text):
        if not self.read_only():
            self.element.name = text or ""

    def sync(self) -> None:
        if element := self.element:
            self.readonly_text = format(element) or gettext("<None>")
            self.notify("editable-text")
            self.icon = icon_name(element)
            self.icon_visible = bool(
                self.icon
                and not isinstance(
                    element, (UML.Parameter, UML.Property, UML.Operation)
                )
            )
            self.attributes = pango_attributes(element)

    def start_editing(self):
        self.editing = True


class RelationshipItem(TreeItem):
    def __init__(self, child_model):
        super().__init__(None)
        self.child_model = child_model
        self.readonly_text = gettext("<Relationships>")
        self.can_edit = False


class Branch:
    def __init__(self):
        self.elements = Gio.ListStore.new(TreeItem.__gtype__)
        self.relationships = Gio.ListStore.new(TreeItem.__gtype__)

    def append(self, element: Element):
        if isinstance(element, UML.Relationship):
            if self.relationships.get_n_items() == 0:
                self.elements.insert(0, RelationshipItem(self.relationships))
            self.relationships.append(TreeItem(element))
        else:
            self.elements.append(TreeItem(element))

    def remove(self, element):
        list_store = (
            self.relationships
            if isinstance(element, UML.Relationship)
            else self.elements
        )
        if (
            index := next(
                (i for i, ti in enumerate(list_store) if ti.element is element),
                None,
            )
        ) is not None:
            list_store.remove(index)

        # Clean up empty relationships node
        if list_store is self.relationships and self.relationships.get_n_items() == 0:
            for i, e in enumerate(self.elements):
                if isinstance(e, RelationshipItem):
                    self.elements.remove(i)
                    break

    def remove_all(self):
        self.relationships.remove_all()
        self.elements.remove_all()

    def changed(self, element: Element):
        list_store = (
            self.relationships
            if isinstance(element, UML.Relationship)
            else self.elements
        )
        if not (
            tree_item := next((ti for ti in list_store if ti.element is element), None)
        ):
            return
        found, index = list_store.find(tree_item)
        if found:
            list_store.items_changed(index, 1, 1)

    def __len__(self):
        return self.elements.get_n_items()

    def __getitem__(self, index):
        return self.elements.get_item(index)

    def __iter__(self):
        yield from self.elements
        yield from self.relationships


def visible(element):
    return isinstance(
        element, (UML.Relationship, UML.NamedElement, Diagram)
    ) and not isinstance(
        element, (UML.InstanceSpecification, UML.OccurrenceSpecification)
    )


def tree_item_sort(a, b, _user_data=None):
    if isinstance(a, RelationshipItem):
        return -1
    if isinstance(b, RelationshipItem):
        return 1
    na = normalize("NFC", a.readonly_text).casefold()
    nb = normalize("NFC", b.readonly_text).casefold()
    return (na > nb) - (na < nb)


class TreeModel:
    def __init__(self):
        super().__init__()
        self.branches: dict[TreeItem | None, Branch] = {None: Branch()}

    @property
    def root(self) -> Gio.ListStore:
        return self.branches[None].elements

    def sync(self, element):
        if visible(element) and (tree_item := self.tree_item_for_element(element)):
            tree_item.sync()

    def child_model(self, item: TreeItem, _user_data=None) -> Gio.ListStore:
        """This method will create branches on demand (lazy)."""
        branches = self.branches
        if item in branches:
            return branches[item].elements
        elif isinstance(item, RelationshipItem):
            return item.child_model
        elif not item.element:
            return None
        elif owned_elements := [
            e
            for e in item.element.ownedElement
            if e.owner is item.element and visible(e)
        ] + (
            [
                e
                for e in item.element.member
                if e.memberNamespace is item.element and not e.owner and visible(e)
            ]
            if isinstance(item.element, UML.Namespace)
            else []
        ):
            new_branch = Branch()
            self.branches[item] = new_branch
            for e in owned_elements:
                new_branch.append(e)
            return new_branch.elements
        return None

    def owner_branch_for_element(
        self, element: Element, former_owner=_no_value
    ) -> Branch | None:
        if (
            owner := (
                element.owner
                or (
                    element.memberNamespace
                    if isinstance(element, UML.NamedElement)
                    else None
                )
            )
            if former_owner is _no_value
            else former_owner
        ) is None:
            return self.branches[None]

        return next(
            (m for ti, m in self.branches.items() if ti and ti.element is owner),
            None,
        )

    def tree_item_for_element(self, element: Element | None) -> TreeItem | None:
        if element is None:
            return None
        if owner_branch := self.owner_branch_for_element(element):
            return next((ti for ti in owner_branch if ti.element is element), None)
        return None

    def add_element(self, element: Element) -> None:
        if (not visible(element)) or self.tree_item_for_element(element):
            return

        if (owner_branch := self.owner_branch_for_element(element)) is not None:
            owner_branch.append(element)
        elif element.owner:
            self.notify_child_model(element.owner)
        elif isinstance(element, UML.NamedElement) and element.memberNamespace:
            self.notify_child_model(element.memberNamespace)

    def remove_element(self, element: Element, former_owner=_no_value) -> None:
        for child in element.ownedElement:
            self.remove_element(child)

        if isinstance(element, UML.Namespace):
            for child in element.member:
                self.remove_element(child)

        # Deal with member relation, but exclude namespace, since it also relates to the owner
        former_namespace = (
            element.memberNamespace
            if (
                former_owner is None
                and isinstance(element, UML.NamedElement)
                and element.memberNamespace
                and element.memberNamespace is not element.namespace
            )
            else None
        )

        if (
            owner_branch := (
                self.owner_branch_for_element(element, former_owner=former_owner)
                or (
                    former_namespace
                    and self.owner_branch_for_element(
                        element, former_owner=former_namespace
                    )
                    or None
                )
            )
        ) is not None:
            owner_branch.remove(element)

            if not len(owner_branch):
                self.remove_branch(owner_branch)

    def remove_branch(self, branch: Branch) -> None:
        tree_item = next(ti for ti, b in self.branches.items() if b is branch)
        if tree_item is None:
            # Do never remove the root branch
            return

        del self.branches[tree_item]

        self.notify_child_model(tree_item.element)

    def notify_child_model(self, element):
        # Only notify the change, the branch is created in child_model()
        owner_tree_item = self.tree_item_for_element(
            element.owner
            or (
                element.memberNamespace
                if isinstance(element, UML.NamedElement)
                else None
            )
        )
        if (
            not self.branches.get(self.tree_item_for_element(element))
            and (owner_branch := self.branches.get(owner_tree_item)) is not None
        ):
            owner_branch.changed(element)

    def clear(self) -> None:
        root = self.branches[None]
        root.remove_all()
        self.branches.clear()
        self.branches[None] = root


def pango_attributes(element):
    attrs = Pango.AttrList.new()
    attrs.insert(
        Pango.attr_weight_new(
            Pango.Weight.BOLD if isinstance(element, Diagram) else Pango.Weight.NORMAL
        )
    )
    attrs.insert(
        Pango.attr_style_new(
            Pango.Style.ITALIC
            if isinstance(element, (UML.Classifier, UML.BehavioralFeature))
            and element.isAbstract
            else Pango.Style.NORMAL
        )
    )
    return attrs



"""Containment connection adapters."""

from gaphor import UML
from gaphor.core.modeling import Diagram, self_and_owners
from gaphor.diagram.connectors import BaseConnector, Connector
from gaphor.diagram.group import group, ungroup
from gaphor.diagram.presentation import ElementPresentation
from gaphor.UML.classes.containment import ContainmentItem
from gaphor.UML.recipes import owner_package


@Connector.register(ElementPresentation, ContainmentItem)
class ContainmentConnect(BaseConnector):
    """Connect a containment relationship to elements that can be nested.

    Head is the container, tail connects to the contained item.
    """

    line: ContainmentItem

    def allow(self, handle, port):
        """In addition to the normal check, both line ends may not be connected
        to the same element."""
        container, contained = self.container_and_contained_element(handle)

        if not container or not contained:
            return True

        if contained in self_and_owners(container):
            return False

        return super().allow(handle, port) and (
            isinstance(container, UML.Package)
            and isinstance(contained, (UML.Type, UML.Package))
            or isinstance(container, UML.Package)
            and isinstance(contained, (Diagram))
            or isinstance(container, UML.Class)
            and isinstance(contained, UML.Classifier)
        )

    def container_and_contained_element(self, handle):
        line = self.line
        opposite = line.opposite(handle)
        connected_to = self.get_connected(opposite)
        if not (connected_to and connected_to.subject):
            return None, None
        elif handle is line.head:
            return self.element.subject, connected_to.subject
        return connected_to.subject, self.element.subject

    def connect(self, handle, port) -> bool:
        container, contained = self.container_and_contained_element(handle)
        return group(container, contained) if container and contained else False  # type: ignore[no-any-return]

    def disconnect(self, handle):
        opposite = self.line.opposite(handle)
        oct = self.get_connected(opposite)
        hct = self.get_connected(handle)

        if hct and oct:
            if oct.subject and hct.subject in oct.subject.ownedElement:
                assert isinstance(hct.subject, (UML.Type, UML.Package))
                ungroup(oct.subject, hct.subject)
                hct.subject.package = owner_package(hct.diagram.owner)
            if hct.subject and oct.subject in hct.subject.ownedElement:
                assert isinstance(oct.subject, (UML.Type, UML.Package))
                ungroup(hct.subject, oct.subject)
                oct.subject.package = owner_package(oct.diagram.owner)

        super().disconnect(handle)



"""The definition for the interactions section of the toolbox."""

from gaphas.item import SE
from gaphas.segment import Segment

from gaphor import UML
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.UML import diagramitems
from gaphor.UML.recipes import owner_package
from gaphor.UML.toolboxconfig import namespace_config


def interaction_config(new_item):
    subject = new_item.subject
    subject.name = new_item.diagram.gettext("New {name}").format(
        name=new_item.diagram.gettext(type(subject).__name__)
    )
    if subject.interaction:
        return

    diagram = new_item.diagram
    package = owner_package(diagram.owner)

    interactions = (
        [i for i in package.ownedType if isinstance(i, UML.Interaction)]
        if package
        else diagram.model.lselect(
            lambda e: isinstance(e, UML.Interaction) and e.package is None
        )
    )
    if interactions:
        interaction = interactions[0]
    else:
        interaction = subject.model.create(UML.Interaction)
        interaction.name = new_item.diagram.gettext("Interaction")
        interaction.package = package

    subject.interaction = interaction


def reflexive_message_config(new_item):
    Segment(new_item, new_item.diagram).split_segment(0, count=3)

    new_item.handles()[1].pos = (40, 0)
    new_item.handles()[2].pos = (40, 20)
    new_item.handles()[-1].pos = (0, 20)
    new_item.horizontal = True
    new_item.orthogonal = True


interactions = ToolSection(
    gettext("Interactions"),
    (
        ToolDef(
            "toolbox-interaction",
            gettext("Interaction"),
            "gaphor-interaction-symbolic",
            "<Shift>N",
            new_item_factory(
                diagramitems.InteractionItem,
                UML.Interaction,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-lifeline",
            gettext("Lifeline"),
            "gaphor-lifeline-symbolic",
            "v",
            new_item_factory(
                diagramitems.LifelineItem,
                UML.Lifeline,
                config_func=interaction_config,
            ),
            handle_index=-1,
        ),
        ToolDef(
            "toolbox-execution-specification",
            gettext("Execution Specification"),
            "gaphor-execution-specification-symbolic",
            None,
            new_item_factory(diagramitems.ExecutionSpecificationItem),
            handle_index=0,
        ),
        ToolDef(
            "toolbox-message",
            gettext("Message"),
            "gaphor-message-symbolic",
            "M",
            new_item_factory(diagramitems.MessageItem),
        ),
        ToolDef(
            "toolbox-reflexive-message",
            gettext("Reflexive message"),
            "gaphor-reflexive-message-symbolic",
            None,
            new_item_factory(
                diagramitems.MessageItem, config_func=reflexive_message_config
            ),
        ),
    ),
)



"""Node item may represent a node or a device UML metamodel classes.

Grouping
========
Node item can group following items

- other nodes, which are represented with Node.nestedNode on UML metamodel
  level
- deployed artifacts using deployment
- components, which are parts of a node acting as structured classifier
  (nodes may have internal structures)

Node item grouping logic is implemented in `gaphor.adapters.grouping`
module.
"""

from gaphor import UML
from gaphor.core.modeling.properties import attribute
from gaphor.core.styling import JustifyContent
from gaphor.diagram.presentation import Classified, ElementPresentation, text_name
from gaphor.diagram.shapes import Box, Text, stroke
from gaphor.diagram.support import represents
from gaphor.UML.classes.stereotype import stereotype_compartments
from gaphor.UML.recipes import stereotypes_str


@represents(UML.Node)
@represents(UML.Device)
class NodeItem(Classified, ElementPresentation):
    """Representation of node or device from UML Deployment package."""

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("children", self.update_shapes)
        self.watch("show_stereotypes", self.update_shapes)
        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype", self.update_shapes)
        self.watch("subject.appliedStereotype.classifier.name")
        self.watch("subject.appliedStereotype.slot", self.update_shapes)
        self.watch("subject.appliedStereotype.slot.definingFeature.name")
        self.watch("subject.appliedStereotype.slot.value", self.update_shapes)
        self.watch("subject[Node].ownedConnector", self.update_shapes)

    show_stereotypes: attribute[int] = attribute("show_stereotypes", int)

    def update_shapes(self, event=None):
        self.shape = Box(
            Box(
                Text(
                    text=lambda: stereotypes_str(
                        self.subject,
                        isinstance(self.subject, UML.Device)
                        and (self.diagram.gettext("device"),)
                        or (),
                    ),
                ),
                text_name(self),
                style={
                    "padding": (4, 4, 4, 4),
                    "justify-content": JustifyContent.START,
                },
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            style={
                "justify-content": JustifyContent.START
                if self.diagram and self.children
                else JustifyContent.CENTER,
            },
            draw=draw_node,
        )


def draw_node(box, context, bounding_box):
    cr = context.cairo

    d = 10
    w = bounding_box.width
    h = bounding_box.height

    cr.rectangle(0, 0, w, h)

    cr.move_to(0, 0)
    cr.line_to(d, -d)
    cr.line_to(w + d, -d)
    cr.line_to(w + d, h - d)
    cr.line_to(w, h)
    cr.move_to(w, 0)
    cr.line_to(w + d, -d)

    stroke(context, fill=True)



from gi.repository import GLib, Gtk


def init_macos_shortcuts():
    def new_shortcut_with_args(shortcut, signal, *args):
        shortcut = Gtk.Shortcut.new(
            trigger=Gtk.ShortcutTrigger.parse_string(shortcut),
            action=Gtk.SignalAction.new(signal),
        )
        if args:
            shortcut.set_arguments(GLib.Variant.new_tuple(*args))
        return shortcut

    def add_move_binding(widget_class, shortcut, step, count):
        widget_class.add_shortcut(
            new_shortcut_with_args(
                shortcut,
                "move-cursor",
                GLib.Variant.new_int32(step),
                GLib.Variant.new_int32(count),
                GLib.Variant.new_boolean(False),
            )
        )

        widget_class.add_shortcut(
            new_shortcut_with_args(
                "|".join(f"<Shift>{s}" for s in shortcut.split("|")),
                "move-cursor",
                GLib.Variant.new_int32(step),
                GLib.Variant.new_int32(count),
                GLib.Variant.new_boolean(True),
            )
        )

    for widget_class in (Gtk.Text, Gtk.TextView):
        for shortcut, signal in (
            ("<Meta>x", "cut-clipboard"),
            ("<Meta>c", "copy-clipboard"),
            ("<Meta>v", "paste-clipboard"),
        ):
            widget_class.add_shortcut(new_shortcut_with_args(shortcut, signal))

        for shortcut, action in (
            ("<Meta>z", "text.undo"),
            ("<Meta><Shift>z", "text.redo"),
        ):
            widget_class.add_shortcut(
                Gtk.Shortcut.new(
                    trigger=Gtk.ShortcutTrigger.parse_string(shortcut),
                    action=Gtk.NamedAction.new(action),
                )
            )

        for shortcut, step, count in (
            ("<Meta>Up|<Meta>KP_Up", Gtk.MovementStep.BUFFER_ENDS, -1),
            ("<Meta>Down|<Meta>KP_Down", Gtk.MovementStep.BUFFER_ENDS, 1),
            ("<Meta>Left|<Meta>KP_Left", Gtk.MovementStep.DISPLAY_LINE_ENDS, -1),
            ("<Meta>Right|<Meta>KP_Right", Gtk.MovementStep.DISPLAY_LINE_ENDS, 1),
            ("<Alt>Left|<Alt>KP_Left", Gtk.MovementStep.WORDS, -1),
            ("<Alt>Right|<Alt>KP_Right", Gtk.MovementStep.WORDS, 1),
        ):
            add_move_binding(widget_class, shortcut, step, count)

    # Gtk.Text

    Gtk.Text.add_shortcut(
        Gtk.Shortcut.new(
            trigger=Gtk.ShortcutTrigger.parse_string("<Meta>a"),
            action=Gtk.CallbackAction.new(lambda self, data: self.select_region(0, -1)),
        )
    )
    Gtk.Text.add_shortcut(
        new_shortcut_with_args(
            "<Meta><Shift>a",
            "move-cursor",
            GLib.Variant.new_int32(Gtk.MovementStep.VISUAL_POSITIONS),
            GLib.Variant.new_int32(0),
            GLib.Variant.new_boolean(False),
        )
    )

    # Gtk.TextView

    Gtk.TextView.add_shortcut(
        new_shortcut_with_args("<Meta>a", "select-all", GLib.Variant.new_boolean(True))
    )
    Gtk.TextView.add_shortcut(
        new_shortcut_with_args(
            "<Meta><Shift>a", "select-all", GLib.Variant.new_boolean(False)
        )
    )



from __future__ import annotations

from gaphor.diagram.group import group, ungroup
from gaphor.UML.uml import Region, State, StateMachine, Transition, Vertex


@group.register(State, Vertex)
@group.register(StateMachine, Vertex)
def state_machine_vertex_group(state_machine: State | StateMachine, vertex: Vertex):
    """Add State to a State Machine, add a Region if necessary."""
    if state_machine.region:
        region = state_machine.region[0]
    else:
        region = state_machine.model.create(Region)
        state_machine.region = region

    vertex.container = region
    return True


@ungroup.register(State, Vertex)
@ungroup.register(StateMachine, Vertex)
def state_machine_vertex_ungroup(state_machine: State | StateMachine, vertex: Vertex):
    if vertex.container and vertex.container in state_machine.region:
        del vertex.container
        return True
    return False


@group.register(Region, Vertex)
@group.register(Region, Transition)
def region_vertex_group(region, vertex_or_transition):
    vertex_or_transition.container = region
    return True


@ungroup.register(Region, Vertex)
@ungroup.register(Region, Transition)
def region_vertex_ungroup(region, vertex_or_transition):
    if vertex_or_transition.container is region:
        del vertex_or_transition.container
        return True
    return False



from gaphor.diagram.group import group, ungroup
from gaphor.UML.uml import (
    Activity,
    ActivityNode,
    ActivityParameterNode,
    ActivityPartition,
)


@group.register(Activity, ActivityNode)
def activity_node_group(activity, node):
    activity.node = node
    return True


@ungroup.register(Activity, ActivityNode)
def activity_node_ungroup(activity, node):
    if node in activity.node:
        activity.node.remove(node)
        return True
    return False


@group.register(Activity, ActivityParameterNode)
def activity_parameter_node_group(activity, node):
    return False


@ungroup.register(Activity, ActivityParameterNode)
def activity_parameter_node_ungroup(activity, node):
    return False


@group.register(ActivityPartition, ActivityNode)
def activity_node_partition_group(partition, node):
    """Group activity nodes within activity partition."""
    if partition.subpartition:
        return False
    partition.node = node
    return True


@ungroup.register(ActivityPartition, ActivityNode)
def activity_node_partition_ungroup(partition, node):
    if node in partition.node:
        partition.node.remove(node)
        return True
    return False



"""Application settings support for Gaphor."""

import logging
from enum import Enum

from gi.repository import Gio

APPLICATION_ID = "org.gaphor.Gaphor"

logger = logging.getLogger(__name__)


class StyleVariant(Enum):
    SYSTEM = 0
    DARK = 1
    LIGHT = 2


class Settings:
    """Gaphor settings."""

    def __init__(self):
        schema_source = Gio.SettingsSchemaSource.get_default()
        self._gio_settings = (
            Gio.Settings.new(APPLICATION_ID)
            if schema_source and schema_source.lookup(APPLICATION_ID, False)
            else None
        )
        if not self._gio_settings:
            logger.warning(
                "Settings schema not found and settings won't be saved, run `poe install-schemas`"
            )

    @property
    def style_variant(self) -> StyleVariant:
        return (
            StyleVariant(self._gio_settings.get_enum("style-variant"))
            if self._gio_settings
            else StyleVariant.SYSTEM
        )

    @style_variant.setter
    def style_variant(self, style_variant: StyleVariant):
        if self._gio_settings:
            self._gio_settings.set_enum("style-variant", style_variant.value)

    def bind_style_variant(self, target, prop):
        # Bind with mapping not supported by PyGObject: https://gitlab.gnome.org/GNOME/pygobject/-/issues/98
        # To bind to a function that can map between guint and a string
        if self._gio_settings:
            style_variant = self._gio_settings.get_enum("style-variant")
            target.set_property(prop, style_variant)

    def style_variant_changed(self, callback):
        if self._gio_settings:

            def _on_changed(_settings, _name):
                callback(self.style_variant)

            self._gio_settings.connect("changed::style-variant", _on_changed)
            callback(self.style_variant)

    @property
    def use_english(self) -> bool:
        return (  # type: ignore[no-any-return]
            self._gio_settings.get_boolean("use-english")
            if self._gio_settings
            else False
        )

    def bind_use_english(self, target, prop):
        if self._gio_settings:
            self._gio_settings.bind(
                "use-english", target, prop, Gio.SettingsBindFlags.DEFAULT
            )


settings = Settings()



from gaphor.diagram.copypaste import copy
from gaphor.UML import ExecutionSpecification, Message
from gaphor.UML.copypaste import copy_named_element


@copy.register
def copy_message(element: Message):
    yield element.id, copy_named_element(element)
    if element.sendEvent:
        yield from copy(element.sendEvent)
    if element.receiveEvent:
        yield from copy(element.receiveEvent)


@copy.register
def copy_execution_specification(element: ExecutionSpecification):
    yield element.id, copy_named_element(element)
    for occurrence in element.executionOccurrenceSpecification:
        yield from copy(occurrence)



"""Sequence and communication diagram messages.

Messages are implemented according to UML 2.1.1 specification.

Implementation Details
======================
Message sort is supported but occurrence specification is not implemented.
This means that model drawn on a diagram is not complete on UML datamodel
level, still it is valid UML diagram (see Lifelines Diagram in UML
specification, page 461).

Reply Messages
--------------
Different sources show that reply message has filled arrow, including
UML 2.0.

UML 2.1.1 specification says that reply message should be drawn with an
open arrow. This is visible on examples in UML 2.0 and UML 2.1.1
specifications.

Asynchronous Signal
--------------------
It is not clear how to draw signals. It is usually drawn with a half-open
arrow.  This approach is used in Gaphor, too.

Delete Message
--------------
Different sources show that delete message has a "X" at the tail.
It does not seem to be correct solution. A "X" should be shown
at the end of lifeline's lifetime instead (see ``lifeline`` module
documentation for more information).

Events
------
Occurrence specification is not implemented, therefore
- no events implemented (i.e. destroy event)
- no message sequence number on communication diagram

Operations
----------
``Lifeline.represents`` attribute is ``None``, so it is not possible to
specify operation (or signal) for a message. Instead, one has to put
operation information in message's name.

See also ``lifeline`` module documentation.
"""

from math import pi

from gaphas.connector import Handle
from gaphas.constraint import constraint

from gaphor import UML
from gaphor.diagram.presentation import LinePresentation, Named, get_center_pos
from gaphor.diagram.shapes import Box, Text, cairo_state, stroke
from gaphor.diagram.support import represents
from gaphor.UML.interactions.lifeline import LifelineItem
from gaphor.UML.recipes import stereotypes_str

PI_2 = pi / 2


@represents(UML.Message)
class MessageItem(Named, LinePresentation[UML.Message]):
    """Message item is drawn on sequence and communication diagrams.

    On communication diagram, message item is decorated with an arrow in
    the middle of a line.

    Multiple messages can be depicted via this one message item.

    Attributes:

    - _is_communication: check if message is on communication diagram
    - _arrow_pos: decorating arrow position
    - _arrow_angle: decorating arrow angle
    """

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=Box(
                Text(
                    text=lambda: stereotypes_str(self.subject),
                ),
                Text(text=lambda: self.subject.name or ""),
            ),
        )
        self.handles()[1].pos = (40, 0)

        self._horizontal_line = constraint(horizontal=[h.pos for h in self.handles()])
        self.diagram.connections.add_constraint(self, self._horizontal_line)

        self._is_communication = False
        self._arrow_pos = 0, 0
        self._arrow_angle = 0

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")

    def load(self, name, value):
        if name == "points":
            self.diagram.connections.remove_constraint(self, self._horizontal_line)
        return super().load(name, value)

    def postload(self):
        super().postload()
        if len(self.handles()) == 2:
            self.diagram.connections.add_constraint(self, self._horizontal_line)

    def insert_handle(self, index: int, handle: Handle) -> None:
        if len(self.handles()) == 2:
            self.diagram.connections.remove_constraint(self, self._horizontal_line)
        super().insert_handle(index, handle)

    def remove_handle(self, handle: Handle) -> None:
        super().remove_handle(handle)
        if len(self.handles()) == 2:
            self.diagram.connections.add_constraint(self, self._horizontal_line)

    def _draw_circle(self, cr):
        """Draw circle for lost/found messages."""
        # method is called by draw_head or by draw_tail methods,
        # so draw in (0, 0))
        cr.set_line_width(0.01)
        cr.arc(0.0, 0.0, 4, 0.0, 2 * pi)
        cr.fill()

    def _draw_arrow(self, cr, half=False, filled=True):
        """Draw an arrow.

        Parameters:

        - half: draw half-open arrow
        - filled: draw filled arrow
        """
        cr.move_to(15, 6)
        cr.line_to(0, 0)
        if not half:
            cr.line_to(15, -6)
        if filled:
            cr.close_path()
            cr.fill_preserve()

    def draw_head(self, context):
        cr = context.cairo
        # no head drawing in case of communication diagram
        if self._is_communication:
            cr.move_to(0, 0)
            return

        cr.move_to(0, 0)

        subject = self.subject
        if subject and subject.messageKind == "found":
            self._draw_circle(cr)
            cr.stroke()

        cr.move_to(0, 0)

    def draw_tail(self, context):
        cr = context.cairo

        # no tail drawing in case of communication diagram
        if self._is_communication:
            cr.line_to(0, 0)
            return

        subject = self.subject

        if subject and subject.messageSort in ("createMessage", "reply"):
            cr.set_dash((7.0, 5.0), 0)

        cr.line_to(0, 0)
        stroke(context, fill=False, dash=False)

        cr.set_dash((), 0)

        if subject:
            w = cr.get_line_width()
            if subject.messageKind == "lost":
                self._draw_circle(cr)
                cr.stroke()

            cr.set_line_width(w)
            half = subject.messageSort == "asynchSignal"
            filled = subject.messageSort in ("synchCall", "deleteMessage")
            self._draw_arrow(cr, half, filled)
        else:
            self._draw_arrow(cr)

        stroke(context, fill=False)

    def _draw_decorating_arrow(self, cr, inverted=False):
        with cairo_state(cr):
            angle: float = self._arrow_angle

            hint = 1 if abs(angle) >= PI_2 and angle != -PI_2 else -1
            if inverted:
                angle += hint * pi

            x, y = self._arrow_pos

            # move to arrow pos and rotate, below we operate in horizontal
            # mode
            cr.translate(x, y)
            cr.rotate(angle)
            # add some padding
            cr.translate(0, 6 * hint)

            # draw decorating arrow
            d = 15
            dr = d - 4
            r = 3
            cr.set_line_width(1.5)
            cr.move_to(-d, 0)
            cr.line_to(d, 0)
            cr.line_to(dr, r)
            cr.move_to(dr, -r)
            cr.line_to(d, 0)
            cr.stroke()

    def draw(self, context):
        super().draw(context)

        # on communication diagram draw decorating arrows for messages and
        # inverted messages
        self._is_communication = self.is_communication()
        if self._is_communication:
            pos, angle = get_center_pos(self.handles())
            self._arrow_pos = pos
            self._arrow_angle = angle
            cr = context.cairo
            self._draw_decorating_arrow(cr)

    def is_communication(self):
        """Check if message is connecting to lifelines on communication
        diagram."""
        c1 = self._connections.get_connection(self.head)
        c2 = self._connections.get_connection(self.tail)
        return (
            c1
            and isinstance(c1.connected, LifelineItem)
            and not c1.connected.lifetime.visible
            and c2
            and isinstance(c2.connected, LifelineItem)
            and not c2.connected.lifetime.visible
        )
